from argparse import ArgumentParser, Namespace
import attr
from pathlib import Path
import subprocess
import sys
import typing

from kgtk.io.kgtkreader import KgtkReader
from kgtk.utils.argparsehelpers import optional_bool

@attr.s(slots=True, frozen=True)
class DocUpdater():
    """
    This class updates KGTK Markdown documents (COMMAND.md files) (specifically, documents in
    mkdocs flavored Markdown).

    - The Usage section, if present, is updated by running:
      - `kgtk COMMAND --help`

      ## Usage
      ```
       usage: ///
      ```

    - The Expert Usage section, if present, is updated by running:
      - `kgtk --expert COMMAND --help`

      ## Expert Usage
      ```
       usage: ///
      ```

    - The Examples section, if present, consists of a number of commands followed by
      - stdout output in a table, or
        - a fenced code block for raw stdout, and
      - an indented code block for stderr output
      - Admonitions may follow the stderr and stdout output.
        - Admonitions may not occur before an stderr output block.
      - Searches for srdoutput tables or blocks, and for stderr output blocks,
        will end at the next line that starts with "#".
      - The Examples section is assumed to run to the end of the file.

      ## Examples

      ### Example 1

      ```bash
      kgtk command
      ```
      | table |

          error output

      ### Example 2

      ```bash
      kgtk command \
           continued
      ```

      ~~~
      raw stdout
      ~~~

          error output

      !!! note
          This is an admonition.


    """


    kgtk_command: str = attr.ib(validator=attr.validators.instance_of(str), default="kgtk")

    process_usage: bool = attr.ib(validator=attr.validators.instance_of(bool), default=True)
    update_usage: bool = attr.ib(validator=attr.validators.instance_of(bool), default=True)

    process_examples: bool = attr.ib(validator=attr.validators.instance_of(bool), default=True)
    update_examples: bool = attr.ib(validator=attr.validators.instance_of(bool), default=True)

    error_file: typing.TextIO = attr.ib(default=sys.stderr)
    verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    very_verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
        
    def readlines(self, md_file: Path)->typing.List[str]:
        with open(md_file, "r") as fd:
            return list(fd) # Read all lines.  Each line (except maybe the last) will end with newline.

    def writelines(self, md_file: Path, lines: typing.List[str]):
        with open(md_file, "w") as fd:
            fd.writelines(lines)

    def find_section(self, section: str, lines: typing.List[str])->int:
        start_of_section: str = "## " + section
        line_number: int
        line: str
        for line_number, line in enumerate(lines):
            if line.startswith(start_of_section):
                return line_number
        return -1

    def find_block(self,
                   lines: typing.List[str],
                   start_idx: int,
                   stop_at_next_section: bool = True,
                   skip_text: bool = True,
                   boundary: str = "```",
                   )->typing.Tuple[int, int]:
        current_idx: int = start_idx
        begin_idx: int = -1
        line: str
        while current_idx < len(lines):
            line = lines[current_idx]
            if stop_at_next_section and line.startswith("#"):
                if self.very_verbose:
                    print("find_block begin search left the section at index %d" % current_idx, file=self.error_file, flush=True)
                return -1, -1
            if line.startswith(boundary):
                current_idx += 1
                begin_idx = current_idx
                break
            if len(line.strip()) < 0 and not skip_text:
                if self.very_verbose:
                    print("find_block begin search found unexpected text at index %d" % current_idx, file=self.error_file, flush=True)
                return -1, -1
            current_idx += 1

        if begin_idx < 0:
            if self.very_verbose:
                print("find_block did not find the beginning of a block.", file=self.error_file, flush=True)
            return -1, -1

        while current_idx < len(lines):
            line = lines[current_idx]
            if stop_at_next_section and line.startswith("#"):
                if self.very_verbose:
                    print("find_block end search left the section at index %d" % current_idx, file=self.error_file, flush=True)
                return -1, -1
            if line.startswith(boundary):
                if self.very_verbose:
                    print("Block found with contents in slice %d:%d" % (begin_idx, current_idx), file=self.error_file, flush=True)
                return begin_idx, current_idx # Return the slice of the contents of the block.
            current_idx += 1
        if self.very_verbose:
            print("find_block did not find the end of a block that began at index %d." % begin_idx, file=self.error_file, flush=True)
        return -1, -1

    def find_table(self,
                   lines: typing.List[str],
                   start_idx: int,
                   stop_at_next_section: bool = False,
                   stop_at_next_block: bool = True,
                   skip_text: bool = False,
                   )->typing.Tuple[int, int]:
        current_idx: int = start_idx
        begin_idx: int = -1
        line: str
        while current_idx < len(lines):
            line = lines[current_idx]
            if stop_at_next_section and line.startswith("#"):
                if self.very_verbose:
                    print("find_table begin search left the section at index %d" % current_idx, file=self.error_file, flush=True)
                return -1, -1
            if stop_at_next_block and line.startswith(("```", "~~~")):
                if self.very_verbose:
                    print("find_table begin search found a block at index %d" % current_idx, file=self.error_file, flush=True)
                return -1, -1
            if line.startswith("|"):
                begin_idx = current_idx
                break
            if len(line.strip()) < 0 and not skip_text:
                if self.very_verbose:
                    print("find_table begin search found unexpected text at index %d" % current_idx, file=self.error_file, flush=True)
                return -1, -1
            current_idx += 1

        if begin_idx < 0:
            if self.very_verbose:
                print("find_table did not find the beginning of a table.", file=self.error_file, flush=True)
            return -1, -1

        while current_idx < len(lines):
            line = lines[current_idx]
            if stop_at_next_section and line.startswith("#"):
                if self.very_verbose:
                    print("find_table end search left the section at index %d" % current_idx, file=self.error_file, flush=True)
                return -1, -1
            if not line.startswith("|"):
                if self.very_verbose:
                    print("Table found with contents in slice %d:%d" % (begin_idx, current_idx), file=self.error_file, flush=True)
                return begin_idx, current_idx # Return the slice of the contents of the block.
            current_idx += 1

        if self.very_verbose:
            print("Table found ending the file with contents in slice %d:%d" % (begin_idx, current_idx), file=self.error_file, flush=True)
        return begin_idx, current_idx # Return the slice of the contents of the block.

    def update_table(self,
                     lines: typing.List[str],
                     table_begin: int,
                     table_end: int,
                     current_idx: int,
                     new_table_lines: typing.List[str])->int:
        nlines: int = len(new_table_lines)

        if self.very_verbose:
            print("Existing table:\n****************\n%s****************" % "".join(lines[table_begin:table_end]), file=self.error_file, flush=True)
            
        if nlines == 0:
            if self.verbose:
                print("Error fetching new table (no output)", file=self.error_file, flush=True)
            current_idx = table_end
        elif not new_table_lines[0].startswith("|"):
            if self.verbose:
                print("Error fetching new table:\n%s" % "".join(new_table_lines), file=self.error_file, flush=True)
            current_idx = table_end
        elif self.update_examples:
            # Replace the old example table with the new example table:
            lines[table_begin:table_end] = new_table_lines
            current_idx = table_end + (nlines - (table_end - table_begin))
        else:
            current_idx = table_end

        return current_idx

    def find_stdout_block(self,
                          lines: typing.List[str],
                          start_idx: int,
                          stop_at_next_section: bool = True,
                          skip_text: bool = True,
                          )->typing.Tuple[int, int]:
        return self.find_block(lines, start_idx, stop_at_next_section, skip_text, boundary='~~~')
        

    def update_stdout_block(self,
                            lines: typing.List[str],
                            stdout_begin: int,
                            stdout_end: int,
                            current_idx: int,
                            new_stdout_lines: typing.List[str])->int:
        nlines: int = len(new_stdout_lines)

        if self.very_verbose:
            print("Existing stdout:\n****************\n%s****************" % "".join(lines[stdout_begin:stdout_end]), file=self.error_file, flush=True)
            
        if nlines == 0:
            if self.verbose:
                print("Error fetching new stdout (no output)", file=self.error_file, flush=True)
            current_idx = stdout_end
        elif self.update_examples:
            # Replace the old example stdout with the new example stdout:
            lines[stdout_begin:stdout_end] = new_stdout_lines
            current_idx = stdout_end + (nlines - (stdout_end - stdout_begin))
        else:
            current_idx = stdout_end

        return current_idx

    def find_code_block(self,
                        lines: typing.List[str],
                        start_idx: int,
                        stop_at_next_section: bool = True,
                        stop_at_next_block: bool = True,
                        stop_at_next_admonition: bool =True,
                        skip_text: bool = False,
                        )->typing.Tuple[int, int]:
        """
        A code block is a set of contiguous indented lines.
        """
        current_idx: int = start_idx
        begin_idx: int = -1
        line: str
        while current_idx < len(lines):
            line = lines[current_idx]
            if stop_at_next_section and line.startswith("#"):
                if self.very_verbose:
                    print("find_code begin search left the section at index %d" % current_idx, file=self.error_file, flush=True)
                return -1, -1
            if stop_at_next_block and line.startswith("```"):
                if self.very_verbose:
                    print("find_code begin search found a block at index %d" % current_idx, file=self.error_file, flush=True)
                return -1, -1
            if stop_at_next_admonition and line.startswith("!!!"):
                if self.very_verbose:
                    print("find_code begin search found an admonition at index %d" % current_idx, file=self.error_file, flush=True)
                return -1, -1
            if line.startswith("    "):
                begin_idx = current_idx
                break
            if len(line.strip()) < 0 and not skip_text:
                if self.very_verbose:
                    print("find_code begin search found unexpected text at index %d" % current_idx, file=self.error_file, flush=True)
                return -1, -1
            current_idx += 1

        if begin_idx < 0:
            if self.very_verbose:
                print("find_code did not find the beginning of a table.", file=self.error_file, flush=True)
            return -1, -1

        while current_idx < len(lines):
            line = lines[current_idx]
            if stop_at_next_section and line.startswith("#"):
                if self.very_verbose:
                    print("find_code end search left the section at index %d" % current_idx, file=self.error_file, flush=True)
                return -1, -1
            if not line.startswith("    "):
                if self.very_verbose:
                    print("Code found with contents in slice %d:%d" % (begin_idx, current_idx), file=self.error_file, flush=True)
                return begin_idx, current_idx # Return the slice of the contents of the block.
            current_idx += 1

        if self.very_verbose:
            print("Code found ending the file with contents in slice %d:%d" % (begin_idx, current_idx), file=self.error_file, flush=True)
        return begin_idx, current_idx # Return the slice of the contents of the block.

    def update_code_block(self,
                    lines: typing.List[str],
                    code_block_begin: int,
                    code_block_end: int,
                    current_idx: int,
                    new_code_lines: typing.List[str])->int:
        nlines: int = len(new_code_lines)

        if self.very_verbose:
            print("Existing code:\n****************\n%s****************" % "".join(lines[code_block_begin:code_block_end]), file=self.error_file, flush=True)
            
        if nlines == 0:
            if self.verbose:
                print("Error fetching new code (no output)", file=self.error_file, flush=True)
            current_idx = code_block_end
        elif not new_code_lines[0].startswith("    "):
            if self.verbose:
                print("Error fetching new code:\n%s" % "".join(new_code_lines), file=self.error_file, flush=True)
            current_idx = code_block_end
        elif self.update_examples:
            # Replace the old example code with the new example code:
            lines[code_block_begin:code_block_end] = new_code_lines
            current_idx = code_block_end + (nlines - (code_block_end - code_block_begin))
        else:
            current_idx = code_block_end

        return current_idx

    def do_usage(self,
                 subcommand: str,
                 lines: typing.List[str],
                 expert_usage: bool = False):

        section_name = "Expert Usage" if expert_usage else "Usage"
        
        usage_section_idx: int = self.find_section(section_name, lines)
        if usage_section_idx < 0:
            if self.verbose:
                print("No %s section found." % section_name, file=self.error_file, flush=True)
            return
        if self.verbose:
            print("%s section found at index %d" % (section_name, usage_section_idx), file=self.error_file, flush=True)

        usage_block_begin: int
        usage_block_end: int
        usage_block_begin, usage_block_end = self.find_block(lines, usage_section_idx + 1)
        if usage_block_begin < 0:
            if self.verbose:
                print("No %s block found." % section_name, file=self.error_file, flush=True)
            return
        if self.very_verbose:
            print("Existing %s:\n****************\n%s****************" % (section_name,
                                                                          "".join(lines[usage_block_begin:usage_block_end])),
                  file=self.error_file, flush=True)

            
        command: str
        if expert_usage:
            command = "%s --expert %s --help" % (self.kgtk_command, subcommand)
        else:
            command = "%s %s --help" % (self.kgtk_command, subcommand)
        if self.verbose:
            print("Getting new %s for %s" % (section_name, repr(command)), file=self.error_file, flush=True)
        new_usage: typing.List[str] = subprocess.getoutput(command).splitlines(keepends=True)
        nlines: int = len(new_usage)
        if nlines > 0:
                new_usage[nlines-1] = new_usage[nlines-1].rstrip('\n') + '\n'
        if self.very_verbose:
            print("new %s:\n****************\n%s****************" % (section_name,
                                                                     "".join(new_usage)),
                  file=self.error_file, flush=True)

        if self.update_usage:
            # Replace the old usage with the new usage:
            lines[usage_block_begin:usage_block_end] = new_usage

    def do_examples(self, subcommand: str, lines: typing.List[str]):
        examples_section_idx: int = self.find_section("Examples", lines)
        if examples_section_idx < 0:
            if self.verbose:
                print("No Examples section found." , file=self.error_file, flush=True)
            return
        if self.verbose:
            print("Examples section found at index %d" % examples_section_idx, file=self.error_file, flush=True)

        current_idx = examples_section_idx + 1

        table_count: int = 0
        missed_table_count: int = 0
        stdout_block_count: int = 0
        missed_stdout_block_count: int = 0
        unexpected_stdout_count: int = 0
        expected_error_count: int = 0
        unexpected_error_count: int = 0
        missed_error_count: int = 0

        while (True):
            # Commands to execute are stored in a fenced code block using "```" as the boundary.
            command_block_begin: int
            command_block_end: int
            command_block_begin, command_block_end = self.find_block(lines, current_idx, stop_at_next_section=False)
            if command_block_begin < 0:
                if self.verbose:
                    print ("%d expected tables processed" % table_count, file=self.error_file, flush=True)
                    print ("%d expected stdout blocks processed" % stdout_block_count, file=self.error_file, flush=True)
                    print ("%d expected errors processed" % expected_error_count, file=self.error_file, flush=True)
                    
                    print ("\n%d missed tables processed" % missed_table_count, file=self.error_file, flush=True)
                    print ("%d missed stdout blocks processed" % missed_stdout_block_count, file=self.error_file, flush=True)
                    print ("%d missed errors processed" % missed_error_count, file=self.error_file, flush=True)
                    print ("%d unexpected stdout outputs processed" % unexpected_stdout_count, file=self.error_file, flush=True)
                    print ("%d unexpected error outputs processed" % unexpected_error_count, file=self.error_file, flush=True)
                return
            if self.very_verbose:
                print("Example command:\n----------\n%s----------" % "".join(lines[command_block_begin:command_block_end]), file=self.error_file, flush=True)
            current_idx = command_block_end + 1

            # Stdout output is processed in a table or in a fenced code block using "~~~" as the boundary..
            command: str = "".join(lines[command_block_begin:command_block_end]).strip()
            table_begin: int
            table_end: int
            table_begin, table_end = self.find_table(lines, current_idx)

            stdout_block_begin: int = -1
            stdout_block_end: int = -1
            if table_begin < 0:
                stdout_block_begin, stdout_block_end = self.find_stdout_block(lines, current_idx)

            if table_begin >= 0:
                command += " / md"
                if self.verbose:
                    print("Getting new table lines for:\n%s" % command, file=self.error_file, flush=True)
            else:
                if self.verbose:
                    print("Not expecting new table lines for:\n%s" % command, file=self.error_file, flush=True)

            result: subprocess.CompletedProcess = subprocess.run(command, capture_output=True, shell=True, text=True)
            new_stdout_lines: typing.List[str] = result.stdout.splitlines(keepends=True)
            nlines: int = len(new_stdout_lines)
            if nlines > 0:
                new_stdout_lines[nlines-1] = new_stdout_lines[nlines-1].rstrip('\n') + '\n'
                if self.verbose:
                    print("%d lines of stdout output." % nlines, file=self.error_file, flush=True)
                if self.very_verbose:
                    print("Stdout output:\n****************\n%s****************" % "".join(new_stdout_lines), file=self.error_file, flush=True)
            elif self.very_verbose:
                print("No stdout output.", file=self.error_file, flush=True)

            if nlines > 0:
                if table_begin >= 0:
                    current_idx = self.update_table(lines, table_begin, table_end, current_idx, new_stdout_lines)
                    table_count += 1
                elif stdout_block_begin >= 0:
                    current_idx = self.update_stdout_block(lines, stdout_block_begin, stdout_block_end, current_idx, new_stdout_lines)
                    stdout_block_count += 1
                else:
                    if self.verbose:
                        print (">>> No table or stdout block found for stdout output", file=self.error_file, flush=True)
                    unexpected_stdout_count += 1
            else:
                if table_begin > 0:
                    if self.verbose:
                        print (">>> No stdout output found for table", file=self.error_file, flush=True)
                    missed_table_count += 1
                elif stdout_block_begin > 0:
                    if self.verbose:
                        print (">>> No stdout output found for stdout block", file=self.error_file, flush=True)
                    missed_stdout_block_count += 1

            new_error_lines: typing.List[str] = result.stderr.splitlines(keepends=True)
            new_error_lines_len: int = len(new_error_lines)
            if new_error_lines_len > 0:
                new_error_lines[new_error_lines_len-1] = new_error_lines[new_error_lines_len-1].rstrip('\n') + '\n'
                if self.verbose:
                    print("%d lines of error output." % new_error_lines_len, file=self.error_file, flush=True)
                if self.very_verbose:
                    print("Error output:\n****************\n%s****************" % "".join(new_error_lines), file=self.error_file, flush=True)
                idx: int
                for idx in range(new_error_lines_len):
                    new_error_lines[idx] = "    " + new_error_lines[idx]
            elif self.very_verbose:
                print("No error output.", file=self.error_file, flush=True)

           # Error output is processed in a Markdown indented code block.
            code_block_begin: int
            code_block_end: int
            code_block_begin, code_block_end = self.find_code_block(lines, current_idx)
            if new_error_lines_len > 0:
                if code_block_begin >= 0:
                    current_idx = self.update_code_block(lines, code_block_begin, code_block_end, current_idx, new_error_lines)
                    expected_error_count += 1
                else:
                    if self.verbose:
                        print (">>> No code block found for error output", file=self.error_file, flush=True)
                    unexpected_error_count += 1
            elif code_block_begin >= 0:
                if self.verbose:
                    print (">>> No error output found for code block", file=self.error_file, flush=True)
                missed_error_count += 1

    def process(self, md_file: Path):
        subcommand: str = md_file.stem
        if self.verbose:
            print("Processing %s for %s %s" % (repr(str(md_file)), self.kgtk_command, subcommand), file=self.error_file, flush=True)
        lines: typing.List[str] = self.readlines(md_file)
        if self.verbose:
            print("Read %d lines." % len(lines), file=self.error_file, flush=True)
        if self.very_verbose:
            line_idx: int
            line: str
            print("============================", file=self.error_file, flush=True)
            for line_idx, line in enumerate(lines):
                print("%d: %s" % (line_idx, line.rstrip()), file=self.error_file, flush=True)
            print("============================", file=self.error_file, flush=True)

        if self.process_usage:
            self.do_usage(subcommand, lines)
            self.do_usage(subcommand, lines, expert_usage=True)

        if self.process_examples:
            self.do_examples(subcommand, lines)

        if self.verbose:
            print("Writing %d lines" % len(lines), file=self.error_file, flush=True)
        self.writelines(md_file, lines)

def main():
    """
    Update the documentation files.
    """
    parser: ArgumentParser = ArgumentParser()
    parser.add_argument("--md-files", dest="md_files", help="The .md files to be updated.", type=Path, nargs='+')

    parser.add_argument("--kgtk-command", dest="kgtk_command", help="The kgtk command (default %(default)s.", type=str, default="kgtk")

    parser.add_argument("--process-usage", dest="process_usage", metavar="optional True|False",
                        help="Process the ## Usage section (default=%(default)s).",
                        type=optional_bool, nargs='?', const=True, default=True)
    
    parser.add_argument("--update-usage", dest="update_usage", metavar="optional True|False",
                        help="Update the ## Usage section (default=%(default)s).",
                        type=optional_bool, nargs='?', const=True, default=True)
    
    parser.add_argument("--process-examples", dest="process_examples", metavar="optional True|False",
                        help="Process the ## Examples section (default=%(default)s).",
                        type=optional_bool, nargs='?', const=True, default=True)
    
    parser.add_argument("--update-examples", dest="update_examples", metavar="optional True|False",
                        help="Update the ## Examples section (default=%(default)s).",
                        type=optional_bool, nargs='?', const=True, default=True)
    
    KgtkReader.add_debug_arguments(parser)

    args: Namespace = parser.parse_args()
    error_file: typing.TextIO = sys.stdout if args.errors_to_stdout else sys.stderr

    if args.show_options:
        print("--md-files %s" % " ".join([repr(str(x)) for x in args.md_files]), file=error_file, flush=True)
        print("--kgtk-command=%s" % repr(args.kgtk_command), file=error_file, flush=True)
        print("--process-usage=%s" % repr(args.process_usage), file=error_file, flush=True)
        print("--update-usage=%s" % repr(args.update_usage), file=error_file, flush=True)
        print("--process-examples=%s" % repr(args.process_examples), file=error_file, flush=True)
        print("--update-examples=%s" % repr(args.update_examples), file=error_file, flush=True)
        print("--verbose=%s" % repr(args.verbose), file=error_file, flush=True)
        print("--very-verbose=%s" % repr(args.very_verbose), file=error_file, flush=True)

    du: DocUpdater = DocUpdater(kgtk_command=args.kgtk_command,
                                process_usage=args.process_usage,
                                update_usage=args.update_usage,
                                process_examples=args.process_examples,
                                update_examples=args.update_examples,
                                error_file=error_file,
                                verbose=args.verbose,
                                very_verbose=args.very_verbose)

    md_file: Path
    for md_file in args.md_files:
        du.process(md_file)


if __name__ == "__main__":
    main()




import re
import json
import gzip
import rfc3986
from typing import List
from etk.etk import ETK
from etk.etk_module import ETKModule
from etk.wikidata.statement import Rank
from etk.knowledge_graph import KGSchema
from etk.wikidata import wiki_namespaces
from kgtk.exceptions import KGTKException
from etk.wikidata.entity import WDItem, WDProperty
from kgtk.io.kgtkreader import KgtkReader

from etk.wikidata.value import (
    Precision,
    Item,
    StringValue,
    TimeValue,
    QuantityValue,
    MonolingualText,
    GlobeCoordinate,
    ExternalIdentifier,
    URLValue
)
from etk.knowledge_graph.node import LiteralType

BAD_CHARS = [":", "&", ",", " ",
             "(", ")", "\'", '\"', "/", "\\", "[", "]", ";", "|"]


class Generator:
    def __init__(self, **kwargs):
        label_set = kwargs.pop("label_set")
        description_set = kwargs.pop("description_set")
        alias_set = kwargs.pop("alias_set")
        n = int(kwargs.pop("n"))
        warning = kwargs.pop("warning")
        log_path = kwargs.pop("log_path")
        self.prop_file = kwargs.pop("prop_file")
        self.read_num_of_lines = 0
        # set sets
        self.set_sets(label_set, description_set, alias_set)
        # column name order_map
        self.order_map = {}
        self.n = n
        self.yyyy_mm_dd_pattern = re.compile(
            r"[12]\d{3}-(0[1-9]|1[0-2])-(0[1-9]|[12]\d|3[01])")
        self.yyyy_pattern = re.compile(r"[12]\d{3}")
        self.quantity_pattern = re.compile(
            r"([\+|\-]?[0-9]+\.?[0-9]*[e|E]?[\-]?[0-9]*)(?:\[([\+|\-]?[0-9]+\.?[0-9]*),([\+|\-]?[0-9]+\.?[0-9]*)\])?([U|Q](?:.*))?")
        self.warning = warning
        if self.warning:
            self.warn_log = open(log_path, "w")
        self.to_append_statement_id = None
        self.corrupted_statement_id = None
        self.to_append_statement = None  # for Json generator
        self.wiki_import_prop_types = set(["wikipedia_sitelink", "language"])
        self.datatype_mapping = {
            # nomenclature from https://w.wiki/Tfn
            "item": Item,
            "WikibaseItem": Item,
            "wikibase-item": Item,

            "time": TimeValue,
            "Time": TimeValue,

            "globe-coordinate": GlobeCoordinate,
            "GlobeCoordinate": GlobeCoordinate,

            "quantity": QuantityValue,
            "Quantity": QuantityValue,

            "monolingualtext": MonolingualText,
            "Monolingualtext": MonolingualText,

            "string": StringValue,
            "String": StringValue,

            "external-identifier": ExternalIdentifier,
            "ExternalId": ExternalIdentifier,
            "external-id": ExternalIdentifier,

            "url": StringValue,  # TODO bug potentially in rdflib
            "Url": StringValue,

            "property": WDProperty,
            "WikibaseProperty": WDProperty,
            "wikibase-property": WDProperty
        }

    def serialize(self):
        raise NotImplemented

    def finalize(self):
        if self.warning:
            self.warn_log.close()
        self.serialize()

    def set_sets(self, label_set: str, description_set: str, alias_set: str):
        self.label_set, self.alias_set, self.description_set = set(label_set.split(",")), set(
            alias_set.split(",")), set(description_set.split(","))

    @staticmethod
    def process_text_string(string: str) -> [str, str]:
        '''
        Language detection is removed from triple generation. The user is responsible for detect the language
        '''
        if len(string) == 0:
            return ["", "en"]
        if "@" in string:
            res = string.split("@")
            text_string = "@".join(res[:-1]).replace('"', "").replace("'", "")
            lang = res[-1].replace('"', '').replace("'", "")
        else:
            text_string = string.replace('"', "").replace("'", "")
            lang = "en"
        return [text_string, lang]

    @staticmethod
    def is_invalid_decimal_string(num_string) -> bool:
        '''
        if a decimal string too small, return True TODO
        '''
        if num_string == None:
            return False
        else:
            if abs(float(num_string)) < 0.0001 and float(num_string) != 0:
                return True
            return False

    @staticmethod
    def is_valid_uri_with_scheme_and_host(uri: str) -> bool:
        '''
        https://github.com/python-hyper/rfc3986/issues/30#issuecomment-461661883
        '''
        try:
            uri = rfc3986.URIReference.from_string(uri)
            rfc3986.validators.Validator().require_presence_of(
                "scheme", "host").check_validity_of("scheme", "host").validate(uri)
            return True
        except:
            return False

    @staticmethod
    def clean_number_string(num: str) -> str:
        from numpy import format_float_positional
        if num == None:
            return None
        else:
            return format_float_positional(float(num), trim="-")

    @staticmethod
    def replace_illegal_string(s: str) -> str:
        '''
        this function serves as the last gate of keeping illegal characters outside of entity creation.
        '''
        for char in BAD_CHARS:
            s = s.replace(char, "_")
        return s


class TripleGenerator(Generator):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        prop_declaration = kwargs.pop("prop_declaration")
        dest_fp = kwargs.pop("dest_fp")
        truthy = kwargs.pop("truthy")
        use_id = kwargs.pop("use_id")
        prefix_path = kwargs.pop("prefix_path")
        self.error_action = kwargs.pop('error_action')
        self.set_prefix(prefix_path)
        self.prop_declaration = prop_declaration
        self.set_properties(self.prop_file)
        if dest_fp.endswith('.gz'):
            self.fp = gzip.open(dest_fp, 'wt')
        self.fp = open(dest_fp, 'w')
        self.truthy = truthy
        self.reset_etk_doc()
        self.serialize_prefix()
        self.use_id = use_id
        self.node1_idx = -1
        self.node2_idx = -1
        self.id_idx = -1
        self.label_idx = -1
        self.kr = None
        self.initialize(kwargs.pop('input_file'))

    def initialize(self, input_file):
        self.input_file = input_file
        self.kr: KgtkReader = KgtkReader.open(self.input_file)
        self.node1_idx = self.kr.get_node1_column_index()
        self.node2_idx = self.kr.get_node2_column_index()
        self.label_idx = self.kr.get_label_column_index()
        self.id_idx = self.kr.get_id_column_index()
        if self.node1_idx == -1:
            raise KGTKException("'node1' column not found")

        if self.node2_idx == -1:
            raise KGTKException("'node2' column not found")

        if self.label_idx == -1:
            raise KGTKException("'label' column not found")

        if self.id_idx == -1:
            raise KGTKException("'id' column not found")

    def process(self):
        input_row_count: int = 2

        if self.prop_declaration:
            for row in self.kr:
                self.read_prop_declaration(row)
        self.kr.close()
        self.kr: KgtkReader = KgtkReader.open(self.input_file)
        for row in self.kr:
            self.entry_point(input_row_count, row)
            input_row_count += 1

        self.finalize()

    def set_prefix(self, prefix_path: str):
        self.prefix_dict = {}
        if prefix_path != "NONE":
            with open(prefix_path, "r") as fp:
                for line_num, edge in enumerate(fp):
                    edge_list = edge.strip("\r\n").split("\t")
                    if line_num == 0:
                        node1_index, node2_index = edge_list.index("node1"), edge_list.index("node2")
                    else:
                        prefix, expand = edge_list[node1_index], edge_list[node2_index]
                        self.prefix_dict[prefix] = expand

    def read_prop_declaration(self, row: List[str]):
        node1, node2, prop, e_id = row[self.node1_idx], row[self.node2_idx], row[self.label_idx], row[self.id_idx]
        if prop == "data_type" or prop == "datatype":
            self.prop_types[node1] = self.datatype_mapping[node2.strip()]

    def set_properties(self, prop_file: str):
        self.prop_types = dict()
        if prop_file == "NONE":
            return

        if prop_file.endswith(".gz"):
            fp = gzip.open(prop_file, 'rt')
        else:
            fp = open(prop_file, "r")

        node1_idx = -1
        node2_idx = -1

        for line in fp:
            vals = line.split('\t')
            vals = [v.strip() for v in vals]
            if 'node1' in vals and 'node2' in vals:
                node1_idx = vals.index('node1')
                node2_idx = vals.index('node2')

            else:
                node1 = vals[node1_idx]
                node2 = vals[node2_idx]
                try:
                    self.prop_types[node1] = self.datatype_mapping[node2.strip()]
                except:
                    self.prop_types[node1] = StringValue
                    if self.error_action == 'log':
                        self.warn_log.write(
                            "DataType {} of node {} is not supported. "
                            "{}'s DataType has been defaulted to StringValue.\n".format(node2, node1, node1)
                        )
                    if self.error_action == 'raise':
                        raise KGTKException("DataType {} of node {} is not supported.\n".format(node2, node1))

    def _node_2_entity(self, node: str):
        '''
        A node can be Qxxx or Pxxx, return the proper entity.
        '''
        if node in self.prop_types:
            entity = WDProperty(node, self.prop_types[node])
        else:
            entity = WDItem(TripleGenerator.replace_illegal_string(node))
        return entity

    def reset_etk_doc(self, doc_id: str = "http://isi.edu/default-ns/projects"):
        """
        reset the doc object and return it. Called at initialization and after outputting triples.
        """
        kg_schema = KGSchema()
        kg_schema.add_schema("@prefix : <http://isi.edu/> .", "ttl")
        self.etk = ETK(kg_schema=kg_schema, modules=ETKModule)
        self.doc = self.etk.create_document({}, doc_id=doc_id)
        for k, v in wiki_namespaces.items():
            if k in self.prefix_dict:
                self.doc.kg.bind(k, self.prefix_dict[k])
            else:
                self.doc.kg.bind(k, v)

    def serialize(self):
        """
        Seriealize the triples. Used a hack to avoid serializing the prefix again.
        """
        docs = self.etk.process_ems(self.doc)
        self.fp.write("\n\n".join(
            docs[0].kg.serialize("ttl").split("\n\n")[1:]))
        self.fp.flush()
        self.reset()

    def serialize_prefix(self):
        """
        This function should be called only once after the doc object is initialized.
        In order to serialize the prefix at the very begining it has to be printed per the change of rdflib 4.2.2->5.0.0
        Relevent issue: https://github.com/RDFLib/rdflib/issues/965
        """
        for k, v in wiki_namespaces.items():
            if k in self.prefix_dict:
                line = "@prefix " + k + ": <" + self.prefix_dict[k] + "> .\n"
            else:
                line = "@prefix " + k + ": <" + v + "> .\n"
            self.fp.write(line)
        self.fp.write("\n")
        self.fp.flush()
        self.reset()

    def reset(self):
        self.to_append_statement_id = None
        self.to_append_statement = None
        self.read_num_of_lines = 0
        self.reset_etk_doc()

    def generate_label_triple(self, node1: str, node2: str) -> bool:
        entity = self._node_2_entity(node1)
        text_string, lang = TripleGenerator.process_text_string(node2)
        entity.add_label(text_string, lang=lang)
        self.doc.kg.add_subject(entity)
        return True

    def generate_description_triple(self, node1: str, node2: str) -> bool:
        entity = self._node_2_entity(node1)
        text_string, lang = TripleGenerator.process_text_string(node2)
        entity.add_description(text_string, lang=lang)
        self.doc.kg.add_subject(entity)
        return True

    def generate_alias_triple(self, node1: str, node2: str) -> bool:
        entity = self._node_2_entity(node1)
        text_string, lang = TripleGenerator.process_text_string(node2)
        entity.add_alias(text_string, lang=lang)
        self.doc.kg.add_subject(entity)
        return True

    def generate_prop_declaration_triple(self, node1: str, node2: str) -> bool:
        # update the known prop_types
        if node1 in self.prop_types:
            if not self.prop_declaration:
                if self.error_action == 'log':
                    self.warn_log.write("IMPORTANT: Duplicated property definition of {} found!."
                                        "Using data type: {} for property {}".format(node1, self.prop_types[node1],
                                                                                     node1))
                if self.error_action == 'raise':
                    raise KGTKException("Duplicated property definition of {} found!".format(node1))
        else:
            self.prop_types[node1] = node2

        prop = WDProperty(node1, self.datatype_mapping[node2])
        self.doc.kg.add_subject(prop)
        return True

    def generate_normal_triple(
            self, node1: str, property: str, node2: str, is_qualifier_edge: bool, e_id: str, line_number: int) -> bool:
        if self.use_id:
            e_id = TripleGenerator.replace_illegal_string(e_id)
        entity = self._node_2_entity(node1)
        edge_type = self.prop_types[property]
        if edge_type == Item:
            object = WDItem(TripleGenerator.replace_illegal_string(node2))
        elif edge_type == WDProperty:
            object = WDProperty(TripleGenerator.replace_illegal_string(node2), self.prop_types[node2])

        elif edge_type == TimeValue:
            if self.yyyy_mm_dd_pattern.match(node2):
                try:
                    dateTimeString = node2
                    object = TimeValue(
                        value=dateTimeString,  # TODO
                        calendar=Item("Q1985727"),
                        precision=Precision.year,
                        time_zone=0,
                    )
                except:
                    return False
            elif self.yyyy_pattern.match(node2):
                try:
                    dateTimeString = node2 + "-01-01"
                    object = TimeValue(
                        value=dateTimeString,  # TODO
                        calendar=Item("Q1985727"),
                        precision=Precision.year,
                        time_zone=0,
                    )
                except:
                    return False
            else:
                try:
                    # TODO, in future, the two cases above will be dropped in principle to comply with the iso format
                    # now it is iso format
                    assert (node2[0] == "^")
                    node2 = node2[1:]  # remove ^
                    if node2.startswith("+"):
                        node2 = node2[1:]
                    dateTimeString, precision = node2.split("/")
                    dateTimeString = dateTimeString[:-1]  # remove Z
                    object = TimeValue(
                        value=dateTimeString,
                        calendar=Item("Q1985727"),
                        precision=precision,
                        time_zone=0,
                    )
                except Exception as e:
                    print(e)
                    return False

        elif edge_type == GlobeCoordinate:
            latitude, longitude = node2[1:].split("/")
            latitude = float(latitude)
            longitude = float(longitude)
            object = GlobeCoordinate(
                latitude, longitude, 0.0001, globe=Item("Q2"))  # earth

        elif edge_type == QuantityValue:
            try:
                res = self.quantity_pattern.match(node2)
                if res == None:
                    self.warn_log.write("Node2 [{}] at line [{}] is not a legal quantity. Skipping it.\n".format(
                        node2, line_number))
                    return False
                res = res.groups()

            except:
                if self.error_action == 'log':
                    self.warn_log.write(
                        "Node2 [{}] at line [{}] is not a legal quantity.\n".format(
                            node2, line_number)
                    )
                if self.error_action == 'raise':
                    raise KGTKException(
                        "Node2 [{}] at line [{}] is not a legal quantity.\n".format(
                            node2, line_number)
                    )

            amount, lower_bound, upper_bound, unit = res

            amount = TripleGenerator.clean_number_string(amount)
            num_type = self.xsd_number_type(amount)

            lower_bound = TripleGenerator.clean_number_string(lower_bound)
            upper_bound = TripleGenerator.clean_number_string(upper_bound)
            if unit != None:
                if upper_bound != None and lower_bound != None:
                    object = QuantityValue(amount, unit=Item(
                        unit), upper_bound=upper_bound, lower_bound=lower_bound, type=num_type)
                else:
                    object = QuantityValue(amount, unit=Item(unit), type=num_type)
            else:
                if upper_bound != None and lower_bound != None:
                    object = QuantityValue(
                        amount, upper_bound=upper_bound, lower_bound=lower_bound, type=num_type)
                else:
                    object = QuantityValue(amount, type=num_type)

        elif edge_type == MonolingualText:
            text_string, lang = TripleGenerator.process_text_string(node2)
            object = MonolingualText(text_string, lang)
        elif edge_type == ExternalIdentifier:
            object = ExternalIdentifier(node2)
        elif edge_type == URLValue:
            if TripleGenerator.is_valid_uri_with_scheme_and_host(node2):
                object = URLValue(node2)
            else:
                return False
        else:
            # treat everything else as stringValue
            object = StringValue(node2)

        if type(object) == WDItem or type(object) == WDProperty:
            self.doc.kg.add_subject(object)

        if is_qualifier_edge:
            # edge: e8 p9 ^2013-01-01T00:00:00Z/11
            # create qualifier edge on previous STATEMENT and return the updated STATEMENT
            self.to_append_statement.add_qualifier(property, object)
            self.doc.kg.add_subject(self.to_append_statement)
        else:
            # edge: q1 p8 q2 e8
            # create brand new property edge and replace STATEMENT
            if self.truthy:
                self.to_append_statement = entity.add_truthy_statement(
                    property, object, statement_id=e_id) if self.use_id else entity.add_truthy_statement(property,
                                                                                                         object)
            else:
                self.to_append_statement = entity.add_statement(
                    property, object, statement_id=e_id) if self.use_id else entity.add_statement(property, object)
            self.doc.kg.add_subject(entity)
        return True

    def entry_point(self, line_number: int, row: List[str]):
        """
        generates a list of two, the first element is the determination of the edge type using corresponding edge type
        the second element is a bool indicating whether this is a valid property edge or qualifier edge.
        Call corresponding downstream functions
        """

        success = True
        node1, node2, prop, e_id = row[self.node1_idx], row[self.node2_idx], row[self.label_idx], row[self.id_idx]
        if line_number == 2:
            # by default a statement edge
            is_qualifier_edge = False
        else:
            if node1 != self.to_append_statement_id and node1 != self.corrupted_statement_id:
                is_qualifier_edge = False
                # also a new statement edge
                if self.read_num_of_lines >= self.n:
                    self.serialize()
            else:
                # qualifier edge or property declaration edge
                is_qualifier_edge = True
                if node1 == self.corrupted_statement_id:
                    self.warn_log.write(
                        "QUALIFIER edge at line [{}] associated of corrupted statement edge of id [{}] dropped.\n".format(
                            line_number, self.corrupted_statement_id
                        )
                    )
                    return
        if prop in self.label_set:
            success = self.generate_label_triple(node1, node2)
        elif prop in self.description_set:
            success = self.generate_description_triple(node1, node2)
        elif prop in self.alias_set:
            success = self.generate_alias_triple(node1, node2)
        elif prop == "data_type" or prop == 'datatype':
            # special edge of prop declaration
            success = self.generate_prop_declaration_triple(node1, node2)
        else:
            if prop in self.prop_types:
                success = self.generate_normal_triple(
                    node1, prop, node2, is_qualifier_edge, e_id, line_number)
            else:
                if self.error_action == 'log':
                    self.warn_log.write("IMPORTANT: property [{}]'s type is unknown at line [{}].\n".format(
                        prop, line_number))
                if self.error_action == 'raise':
                    raise KGTKException(
                        "property [{}]'s type is unknown at line [{}].\n".format(
                            prop, line_number)
                    )
        if (not success):
            if not is_qualifier_edge:
                self.warn_log.write(
                    "CORRUPTED_STATEMENT edge at line: [{}] with edge id [{}].\n".format(
                        line_number, e_id))
                self.corrupted_statement_id = e_id
            else:
                self.warn_log.write(
                    "CORRUPTED_QUALIFIER edge at line: [{}] with edge id [{}].\n".format(
                        line_number, e_id))

        else:
            self.read_num_of_lines += 1
            if not is_qualifier_edge:
                self.to_append_statement_id = e_id

    @staticmethod
    def xsd_number_type(num):
        if isinstance(num, float) and 'e' in str(num).lower():
            return LiteralType.double
        return LiteralType.decimal


class JsonGenerator(Generator):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_declaration = kwargs.pop("prop_declaration")
        self.output_prefix = kwargs.pop("output_prefix")
        self.has_rank = kwargs.pop("has_rank")
        self.error_action = kwargs.pop('error_action')
        self.file_num = 0
        # this data_type mapping is to comply with the SQID UI parsing requirements
        self.datatype_mapping = {
            "item": "wikibase-item",
            "WikibaseItem": "wikibase-item",
            "wikibase-item": "wikibase-item",

            "property": "wikibase-item",
            "WikibaseProperty": "wikibase-item",
            "wikibase-property": "wikibase-item",

            "time": "time",
            "Time": "time",

            "globe-coordinate": "globe-coordinate",
            "GlobeCoordinate": "globe-coordinate",

            "quantity": "quantity",
            "Quantity": "quantity",

            "monolingualtext": "monolingualtext",
            "Monolingualtext": "monolingualtext",

            "string": "string",
            "String": "string",

            "external-identifier": "external-id",
            "ExternalId": "external-id",
            "external-id": "external-id",

            "url": "url",
            "Url": "url"
        }
        self.set_properties(self.prop_file)
        self.set_json_dict()
        self.previous_qnode = None
        self.node1_idx = -1
        self.node2_idx = -1
        self.id_idx = -1
        self.label_idx = -1
        self.kr = None
        self.initialize(kwargs.pop('input_file'))

    def initialize(self, input_file):
        self.input_file = input_file
        self.kr: KgtkReader = KgtkReader.open(self.input_file)
        self.node1_idx = self.kr.get_node1_column_index()
        self.node2_idx = self.kr.get_node2_column_index()
        self.label_idx = self.kr.get_label_column_index()
        self.id_idx = self.kr.get_id_column_index()
        if hasattr(self, 'has_rank') and self.has_rank == True:
            rank_index = self.kr.get_node1_column_index('rank')
            self.rank_idx = rank_index

    def process(self):
        input_row_count: int = 2

        if self.prop_declaration:
            for row in self.kr:
                self.read_prop_declaration(row)
        self.kr.close()
        self.kr: KgtkReader = KgtkReader.open(self.input_file)
        for row in self.kr:
            self.entry_point(input_row_count, row)
            input_row_count += 1

        self.finalize()

    def entry_point(self, line_number, row):
        self.read_num_of_lines += 1

        node1 = row[self.node1_idx]
        node2 = row[self.node2_idx]
        prop = row[self.label_idx]
        e_id = row[self.id_idx]
        if self.has_rank:
            rank = row[self.rank_idx]
        else:
            rank = "normal"  # TODO default rank

        # property declaration
        if prop == "data_type":
            if self.prop_declaration:
                self.prop_types[node1] = self.datatype_mapping[node2.strip()]
                return
            else:
                self.warn_log.write(
                    "CORRUPTED_STATEMENT property declaration edge at line: [{}] with edge id [{}].\n".format(
                        line_number, e_id))

        # add qualifier logic
        if line_number == 2:
            self.previous_qnode = node1
            is_qualifier_edge = False
        else:
            if node1 != self.to_append_statement_id and node1 != self.corrupted_statement_id:
                is_qualifier_edge = False
                # partial serialization
                if self.read_num_of_lines >= self.n:
                    if self.previous_qnode != node1:
                        self.serialize()
                    # update previous qnode to avoid breaking continuous same-qnode statements into two jsonl files
                self.previous_qnode = node1
            else:
                is_qualifier_edge = True
                if node1 == self.corrupted_statement_id:
                    self.warn_log.write(
                        "QUALIFIER edge at line [{}] associated with corrupted statement edge of id [{}] dropped.\n".format(
                            line_number, self.corrupted_statement_id)
                    )

        # update info_json_dict
        if not is_qualifier_edge:
            if node1 in self.prop_types:
                success = self.update_misc_json_dict_info(node1, line_number, self.prop_types[node1])
            else:
                success = self.update_misc_json_dict_info(node1, line_number, None)
            assert (success)

        if prop in self.prop_types:
            success = self.update_misc_json_dict_info(prop, line_number, self.prop_types[prop])
            assert (success)
            if self.prop_types[prop] == "wikibase-item":
                success = self.update_misc_json_dict_info(node2, line_number, None)
                assert (success)

        # update label_json_dict
        if prop in self.label_set:
            success = self.update_misc_json_dict(node1, prop, node2, line_number, rank, "label")
            assert (success)
            return

        # update alias and descriptions
        if prop in self.description_set:
            success = self.update_misc_json_dict(node1, prop, node2, line_number, rank, "description")
            assert (success)
            return

        if prop in self.alias_set:
            success = self.update_misc_json_dict(node1, prop, node2, line_number, rank, "alias")
            assert (success)
            return

            # normal update for claims & qualifiers
        if is_qualifier_edge:
            success = self.update_misc_json_dict(node1, prop, node2, line_number, rank, "qualifier")
        else:
            success = self.update_misc_json_dict(node1, prop, node2, line_number, rank, "statement")

        if (not success):
            if not is_qualifier_edge:
                if self.error_action == 'log':
                    self.warn_log.write(
                        "CORRUPTED_STATEMENT edge at line: [{}] with edge id [{}].\n".format(
                            line_number, e_id))

                    self.corrupted_statement_id = e_id
                if self.error_action == 'raise':
                    raise KGTKException("CORRUPTED_STATEMENT edge at line: [{}] with edge id [{}].\n".format(
                        line_number, e_id))
            else:
                if self.error_action == 'log':
                    self.warn_log.write(
                        "CORRUPTED_QUALIFIER edge at line: [{}] with edge id [{}].\n".format(
                            line_number, e_id))
                if self.error_action == 'raise':
                    raise KGTKException(
                        "CORRUPTED_QUALIFIER edge at line: [{}] with edge id [{}].\n".format(
                            line_number, e_id))
        else:
            # success
            if not is_qualifier_edge:
                self.to_append_statement_id = e_id
                self.to_append_statement = [node1, prop]  # path in dictionary for adding future qualifiers

    def init_entity_in_json(self, node: str):
        self.misc_json_dict[node] = {}
        self.misc_json_dict[node]["labels"] = {}
        self.misc_json_dict[node]["descriptions"] = {}
        self.misc_json_dict[node]["aliases"] = {}
        self.misc_json_dict[node]["claims"] = {}
        self.misc_json_dict[node]["sitelinks"] = {}
        if node not in self.prop_types:
            label_type = "item"
        else:
            label_type = "property"
            label_datatype = self.prop_types[node]
            self.misc_json_dict[node]["datatype"] = label_datatype
        self.misc_json_dict[node]["type"] = label_type
        self.misc_json_dict[node]["id"] = node

    def update_misc_json_dict_info(self, node: str, line_number: int, data_type=None):
        if node not in self.misc_json_dict:
            self.init_entity_in_json(node)

        if node.startswith("Q"):
            self.misc_json_dict[node].update(
                {
                    "pageid": -1,
                    "ns": -1,
                    "title": node,
                    "lastrevid": "2000-01-01T00:00:00Z",
                    "type": "item",
                    "id": node}
            )
        elif node.startswith("P"):
            self.misc_json_dict[node].update(
                {
                    "pageid": -1,
                    "ns": -1,
                    "title": "Property:" + node,
                    "lastrevid": "2000-01-01T00:00:00Z",
                    "type": "property",
                    "datatype": data_type,
                    "id": node}
            )
        else:
            if self.error_action == 'log':
                self.warn_log.write(
                    "node [{}] at line [{}] is neither an entity nor a property.\n".format(node, line_number))
            if self.error_action == 'raise':
                raise KGTKException(
                    "node [{}] at line [{}] is neither an entity nor a property.\n".format(node, line_number))
        return True

    def update_misc_json_dict(self, node1: str, prop: str, node2: str, line_number: int, rank: str, field: str):
        if node1 not in self.misc_json_dict and field != "qualifier":
            self.init_entity_in_json(node1)

        if field == "label":
            label_text, lang = JsonGenerator.process_text_string(node2)
            temp_des_dict = {lang: {"languange": lang, "value": label_text}}
            self.misc_json_dict[node1]["labels"].update(temp_des_dict)
            return True

        if field == "description":
            description_text, lang = JsonGenerator.process_text_string(node2)
            temp_des_dict = {lang: {"languange": lang, "value": description_text}}
            self.misc_json_dict[node1]["descriptions"].update(temp_des_dict)
            return True

        if field == "alias":
            alias_text, lang = JsonGenerator.process_text_string(node2)
            temp_alias_dict = {lang: {"languange": lang, "value": alias_text}}
            if lang in self.misc_json_dict[node1]["aliases"]:
                self.misc_json_dict[node1]["aliases"][lang].append(temp_alias_dict)
            else:
                self.misc_json_dict[node1]["aliases"][lang] = [temp_alias_dict]
            return True

        if field == "statement":
            is_qualifier_edge = False
        elif field == "qualifier":
            is_qualifier_edge = True

        if prop not in self.prop_types:
            if prop in self.wiki_import_prop_types:
                self.warn_log.write(
                    "Property {} created by wikidata json dump at line {} is skipped.\n".format(prop, line_number))
                return True
            else:
                raise KGTKException("property {} at line {} is not defined.".format(prop, line_number))

        if not is_qualifier_edge:
            if prop not in self.misc_json_dict[node1]["claims"]:
                self.misc_json_dict[node1]["claims"][prop] = []
        try:
            if self.prop_types[prop] == "wikibase-item":
                object = self.update_misc_json_dict_item(node1, prop, node2, rank, is_qualifier_edge)
            elif self.prop_types[prop] == "time":
                # print("matched date format yyyy-mm-dd",node1,prop,node2,is_qualifier_edge)
                object = self.update_misc_json_dict_time(node1, prop, node2, rank, is_qualifier_edge)
            elif self.prop_types[prop] == "globe-coordinate":
                object = self.update_misc_json_dict_coordinate(node1, prop, node2, rank, is_qualifier_edge)
            elif self.prop_types[prop] == "quantity":
                object = self.update_misc_json_dict_quantity(node1, prop, node2, rank, is_qualifier_edge)
            elif self.prop_types[prop] == "monolingualtext":
                object = self.update_misc_json_dict_monolingualtext(node1, prop, node2, rank, is_qualifier_edge)
            elif self.prop_types[prop] == "string":
                object = self.update_misc_json_dict_string(node1, prop, node2, rank, is_qualifier_edge)
            elif self.prop_types[prop] == "external-id":
                object = self.update_misc_json_dict_external_id(node1, prop, node2, rank, is_qualifier_edge)
            elif self.prop_types[prop] == "url":
                object = self.update_misc_json_dict_url(node1, prop, node2, rank, is_qualifier_edge)
            else:
                if self.error_action == 'log':
                    self.warn_log.write("property tyepe {} of property {} at line {} is not defined."
                                        .format(self.prop_types[prop], prop, line_number))
                if self.error_action == 'raise':
                    raise KGTKException("property tyepe {} of property {} at line {} is not defined."
                                        .format(self.prop_types[prop], prop, line_number))

            if not object:
                if self.warning:
                    self.warn_log.write("edge creation error at line [{}].\n".format(line_number))
                return False

            # process object
            if is_qualifier_edge:
                # update qualifier edge
                if prop in self.misc_json_dict[self.to_append_statement[0]]["claims"][self.to_append_statement[1]][-1][
                    "qualifiers"]:
                    self.misc_json_dict[self.to_append_statement[0]]["claims"][self.to_append_statement[1]][-1][
                        "qualifiers"][prop].append(object)
                else:
                    self.misc_json_dict[self.to_append_statement[0]]["claims"][self.to_append_statement[1]][-1][
                        "qualifiers"][prop] = [object]
                if prop not in (
                        self.misc_json_dict[self.to_append_statement[0]]["claims"][self.to_append_statement[1]][-1][
                            "qualifiers-order"]):
                    self.misc_json_dict[self.to_append_statement[0]]["claims"][self.to_append_statement[1]][-1][
                        "qualifiers-order"].append(prop)
            else:
                self.misc_json_dict[node1]["claims"][prop].append(object)
            return True

        except:
            raise KGTKException("illegal edge at line {}.".format(line_number))

    def update_misc_json_dict_item(self, node1: str, prop: str, node2: str, rank: str, is_qualifier_edge: bool):
        if not is_qualifier_edge:
            temp_item_dict = {
                "mainsnak": {
                    "snaktype": "value",
                    "property": prop,
                    "hash": "",
                    "datavalue": {
                        "value": {
                            "entity-type": "item",
                            "numeric-id": 0,
                            "id": node2
                        },
                        "type": "wikibase-entityid"
                    },
                    "datatype": "wikibase-item"
                },
                "type": "statement",
                "id": node1 + prop + node2,
                "rank": rank,  # TODO
                "references": [],
                "qualifiers": {},
                "qualifiers-order": []
            }
        else:
            temp_item_dict = {
                "snaktype": "value",
                "property": prop,
                "hash": "",
                "datavalue": {
                    "value": {
                        "entity-type": "item",
                        "numeric-id": 0,
                        "id": node2
                    },
                    "type": "wikibase-entityid"
                },
                "datatype": "wikibase-item"
            }
        return temp_item_dict

    def update_misc_json_dict_time(self, node1: str, prop: str, node2: str, rank: str, is_qualifier_edge: bool):
        if self.yyyy_mm_dd_pattern.match(node2):
            time_string = node2 + "-00-00T00:00:00Z"
            precision = 11
        elif self.yyyy_pattern.match(node2):
            time_string = node2 + "-01-01T00:00:00Z"
            precision = 9
        else:
            try:
                time_string, precision = node2.split("/")
                if time_string.startswith("^"):
                    time_string = time_string[1:]
                if time_string.startswith("+"):
                    time_string = time_string[1:]
                precision = int(precision)
            except:
                return None
        if not is_qualifier_edge:
            temp_time_dict = {
                "mainsnak": {
                    "snaktype": "value",
                    "property": prop,
                    "hash": "",
                    "datavalue": {
                        "value": {
                            "time": time_string,
                            "timezone": 0,
                            "before": 0,
                            "after": 0,
                            "precision": precision,
                            "calendarmodel": "http://www.wikidata.org/entity/Q1985727"
                        },
                        "type": "time"
                    },
                    "datatype": "time"
                },
                "type": "statement",
                "id": node1 + prop + node2,
                "rank": rank,  # TODO
                "references": [],
                "qualifiers": {},
                "qualifiers-order": []
            }
        else:
            temp_time_dict = {
                "snaktype": "value",
                "property": prop,
                "hash": "",
                "datavalue": {
                    "value": {
                        "time": time_string,
                        "timezone": 0,
                        "before": 0,
                        "after": 0,
                        "precision": precision,
                        "calendarmodel": "http://www.wikidata.org/entity/Q1985727"
                    },
                    "type": "time"
                },
                "datatype": "time"
            }
        return temp_time_dict

    def update_misc_json_dict_coordinate(self, node1: str, prop: str, node2: str, rank: str, is_qualifier_edge: bool):
        try:
            latitude, longitude = node2[1:].split("/")
            latitude = float(latitude)
            longitude = float(longitude)
        except:
            return None
        if not is_qualifier_edge:
            temp_coordinate_dict = {
                "mainsnak": {
                    "snaktype": "value",
                    "property": prop,
                    "hash": "",
                    "datavalue": {
                        "value": {
                            "latitude": latitude,
                            "longitude": longitude,
                            "altitude": None,
                            "precision": 0.01,  # TODO
                            "globe": "http://www.wikidata.org/entity/Q2"
                        },
                        "type": "globecoordinate"
                    },
                    "datatype": "globecoordinate"
                },
                "type": "statement",
                "id": node1 + prop + node2,
                "rank": rank,
                "references": [],
                "qualifiers": {},
                "qualifiers-order": []
            }
        else:
            temp_coordinate_dict = {
                "snaktype": "value",
                "property": prop,
                "hash": "",
                "datavalue": {
                    "value": {
                        "latitude": latitude,
                        "longitude": longitude,
                        "altitude": None,
                        "precision": 0.01,  # TODO
                        "globe": "http://www.wikidata.org/entity/Q2"
                    },
                    "type": "globecoordinate"
                },
                "datatype": "globecoordinate"
            }
        return temp_coordinate_dict

    def update_misc_json_dict_quantity(self, node1: str, prop: str, node2: str, rank: str, is_qualifier_edge: bool):
        try:
            res = self.quantity_pattern.match(node2).groups()
            amount, lower_bound, upper_bound, unit = res
            amount = JsonGenerator.clean_number_string(amount)
            lower_bound = JsonGenerator.clean_number_string(lower_bound)
            upper_bound = JsonGenerator.clean_number_string(upper_bound)
            unit = "http://www.wikidata.org/entity/" + unit if unit != None else None
        except:
            return None
        if not is_qualifier_edge:
            temp_quantity_dict = {
                "mainsnak": {
                    "snaktype": "value",
                    "property": prop,
                    "hash": "",
                    "datavalue": {
                        "value": {
                            "amount": amount,
                            "unit": unit,
                            "lowerBound": lower_bound,
                            "UpperBound": upper_bound
                        },
                        "type": "quantity"
                    },
                    "datatype": "quantity"
                },
                "type": "statement",
                "id": node1 + prop + node2,
                "rank": rank,
                "references": [],
                "qualifiers": {},
                "qualifiers-order": []
            }
        else:
            temp_quantity_dict = {
                "snaktype": "value",
                "property": prop,
                "hash": "",
                "datavalue": {
                    "value": {
                        "amount": amount,
                        "unit": unit,
                        "lowerBound": lower_bound,
                        "UpperBound": upper_bound
                    },
                    "type": "quantity"
                },
                "datatype": "quantity"
            }
        return temp_quantity_dict

    def update_misc_json_dict_monolingualtext(self, node1: str, prop: str, node2: str, rank: str,
                                              is_qualifier_edge: bool):
        text_string, lang = JsonGenerator.process_text_string(node2)
        if not is_qualifier_edge:
            temp_mono_dict = {
                "mainsnak": {
                    "snaktype": "value",
                    "property": prop,
                    "hash": "",
                    "datavalue": {
                        "value": {
                            "text": text_string,
                            "language": lang
                        },
                        "type": "monolingualtext"
                    },
                    "datatype": "monolingualtext"
                },
                "type": "statement",
                "id": node1 + prop + node2,
                "rank": rank,
                "references": [],
                "qualifiers": {},
                "qualifiers-order": []
            }
        else:
            temp_mono_dict = {
                "snaktype": "value",
                "property": prop,
                "hash": "",
                "datavalue": {
                    "value": {
                        "text": text_string,
                        "language": lang
                    },
                    "type": "monolingualtext"
                },
                "datatype": "monolingualtext"
            }
        return temp_mono_dict

    def update_misc_json_dict_string(self, node1: str, prop: str, node2: str, rank: str, is_qualifier_edge: bool):
        string, lang = JsonGenerator.process_text_string(node2)
        if not is_qualifier_edge:
            temp_string_dict = {
                "mainsnak": {
                    "snaktype": "value",
                    "property": prop,
                    "hash": "",
                    "datavalue": {"value": string, "type": "string"},
                    "datatype": "string"
                },
                "type": "statement",
                "id": node1 + prop + node2,
                "rank": rank,
                "references": [],
                "qualifiers": {},
                "qualifiers-order": []
            }
        else:
            temp_string_dict = {
                "snaktype": "value",
                "property": prop,
                "hash": "",
                "datavalue": {"value": string, "type": "string"},
                "datatype": "string"
            }
        return temp_string_dict

    def update_misc_json_dict_external_id(self, node1: str, prop: str, node2: str, rank: str, is_qualifier_edge: bool):

        if not is_qualifier_edge:
            temp_e_id_dict = {"mainsnak": {
                "snaktype": "value",
                "property": prop,
                "hash": "",
                "datavalue": {"value": node2, "type": "string"},
                "datatype": "external-id"
            },
                "type": "statement",
                "id": node1 + prop + node2,
                "rank": rank,
                "references": [],
                "qualifiers": {},
                "qualifiers-order": []
            }
        else:
            temp_e_id_dict = {
                "snaktype": "value",
                "property": prop,
                "hash": "",
                "datavalue": {"value": node2, "type": "string"},
                "datatype": "external-id"
            }
        return temp_e_id_dict

    def update_misc_json_dict_url(self, node1: str, prop: str, node2: str, rank: str, is_qualifier_edge: bool):
        if not is_qualifier_edge:
            temp_url_dict = {
                "mainsnak": {
                    "snaktype": "value",
                    "property": prop,
                    "hash": "",
                    "datavalue": {
                        "value": node2,
                        "type": "string"
                    },
                    "datatype": "url"
                },
                "type": "statement",
                "id": node1 + prop + node2,
                "rank": rank,
                "references": [],
                "qualifiers": {},
                "qualifiers-order": []
            }
        else:
            temp_url_dict = {
                "snaktype": "value",
                "property": prop,
                "hash": "",
                "datavalue": {
                    "value": node2,
                    "type": "string"
                },
                "datatype": "url"
            }
        return temp_url_dict

    def read_prop_declaration(self, row: List[str]):
        node1, node2, prop, e_id = row[self.node1_idx], row[self.node2_idx], row[self.label_idx], row[self.id_idx]
        if prop == "data_type" or prop == "datatype":
            self.prop_types[node1] = self.datatype_mapping[node2.strip()]

    def set_properties(self, prop_file: str):
        self.prop_types = {}
        if prop_file == "NONE":
            return
        with open(prop_file, "r") as fp:
            props = fp.readlines()
        for line in props[1:]:
            node1, _, node2, = line.split("\t")
            try:
                self.prop_types[node1] = self.datatype_mapping[node2.strip()]
            except:
                raise KGTKException(
                    "DataType {} of node {} is not supported.\n".format(
                        node2, node1
                    )
                )

    def set_json_dict(self):
        self.misc_json_dict = {}
        # self.label_json_dict = {}
        # self.info_json_dict = {}

    def serialize(self):
        '''
        serialize the dictionaries.
        '''
        with open("{}{}.jsonl".format(self.output_prefix, self.file_num), "w") as fp:
            for key, value in self.misc_json_dict.items():
                json.dump({key: value}, fp)
                fp.write("\n")
        self.file_num += 1
        self.reset()

    def reset(self):
        self.set_json_dict()
        self.read_num_of_lines = 0
        self.to_append_statement_id = None
        self.to_append_statement = None



"""
Import an ATOMIC file to KGTK.

"""

import sys
from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Import ATOMIC into KGTK.' 
    }

def add_arguments(parser: KGTKArgumentParser):
    """
    Parse arguments
    Args:
            parser (argparse.ArgumentParser)
    """
    parser.add_input_file(positional=True)
    parser.add_output_file()

def run(input_file: KGTKFiles, output_file: KGTKFiles):

    # import modules locally
    import sys # type: ignore
    from kgtk.exceptions import kgtk_exception_auto_handler, KGTKException
    import csv
    import re
    import json
    from pathlib import Path
    from string import Template
    import pandas as pd
    from kgtk.kgtkformat import KgtkFormat
    from kgtk.io.kgtkwriter import KgtkWriter

    def make_node(x):
        und_x=x.replace(' ', '_')
        pref_und_x='at:%s' % und_x
        return pref_und_x

    def remove_people_mentions(event):
        e=event.replace('personx', '').strip()
        e=e.replace('persony', '').strip()
        e=e.replace('person x', '').strip()
        e=e.replace('person y', '').strip()
        e=e.replace('the ___', '')
        e=e.replace('___', '')
        e=e.replace("'s", '')
        e=e.replace('to y', '')
        return e.strip()


    def produce_node_labels(event):
        if '\t' in event:
            event=event.split('\t')[0]
        e1=event.lower()
        e1=e1.rstrip('.').strip()
        e2=remove_people_mentions(e1)
        while '  ' in e2:
            e2=e2.replace('  ', ' ')
        if e1!=e2 and e2:
            return '|'.join([KgtkFormat.stringify(e1),KgtkFormat.stringify(e2)])
        else:
            return KgtkFormat.stringify(e1)

    def produce_rel_label(rel):
        mapping={
                    'xAttr': 'person x has attribute',
                    'oAttr': 'others have attribute',
                    'xReact': 'person x feels',
                    'oReact': 'others feel',
                    'xIntent': 'person x wants',
                    'xWant': 'person x wants',
                    'oWant': 'others want',
                    'xNeed': 'person x needs',
                    'xEffect': 'effect on person x',
                    'oEffect': 'the effect on others'
                }
        return KgtkFormat.stringify(mapping[rel])

    try:

        filename: Path = KGTKArgumentParser.get_input_file(input_file)

        out_columns=['node1', 'relation', 'node2', 'node1;label', 'node2;label','relation;label', 'relation;dimension', 'source', 'sentence']

        output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)
        ew: KgtkWriter = KgtkWriter.open(out_columns,
                                         output_kgtk_file,
                                         #mode=input_kr.mode,
                                         require_all_columns=False,
                                         prohibit_extra_columns=True,
                                         fill_missing_columns=True,
                                         gzip_in_parallel=False,
                                         #verbose=self.verbose,
                                         #very_verbose=self.very_verbose
                                         )


        df = pd.read_csv(filename,index_col=0)
        df.iloc[:,:9] = df.iloc[:,:9].apply(lambda col: col.apply(json.loads))

        df.drop(df.columns[len(df.columns)-1], axis=1, inplace=True)
        df.drop(df.columns[len(df.columns)-1], axis=1, inplace=True)

        for event, row in df.iterrows():
            event_label=produce_node_labels(event)

            first_event_label=KgtkFormat.unstringify(event_label.split('|')[0] if '|' in event_label else event_label)
            n1=make_node(first_event_label)
            for c in df.columns:
                for v in row[c]:
                    if v=='none': continue
                    value_label=produce_node_labels(v)
                    first_value_label=KgtkFormat.unstringify(value_label.split('|')[0] if '|' in value_label else value_label)
                    n2=make_node(first_value_label)

                    rel_label=produce_rel_label(c)

                    sentence=''

                    relation=make_node(c)

                    this_row=[n1, relation, n2, event_label, value_label, rel_label, '', KgtkFormat.stringify('AT'), sentence]
                    ew.write(this_row)

        # Clean up.
        ew.close()

    except Exception as e:
        raise KGTKException('Error: ' + str(e))



"""Copy records from the first KGTK file to the output file,
adding ID values.

TODO: Need KgtkWriterOptions
"""

from argparse import Namespace, SUPPRESS

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Copy a KGTK file, adding ID values.',
        'description': 'Copy a KGTK file, adding ID values.\n\n' +
        'The `--overwrite-id` option can be used to replace existing ID values in the ID column.\n' +
        'It does not update instances of the same ID in other columns, such as node1, elsewhere in the file.\n\n' +
        'Several ID styles are supported.\n\n' +
        'Additional options are shown in expert help.\nkgtk --expert add-id --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    # import modules locally
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.reshape.kgtkidbuilder import KgtkIdBuilder, KgtkIdBuilderOptions
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_input_file(positional=True)
    parser.add_output_file()

    KgtkIdBuilderOptions.add_arguments(parser, expert=True) # Show all the options.
    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_file: KGTKFiles,
        output_file: KGTKFiles,

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from pathlib import Path
    import sys
    import typing
    
    from kgtk.exceptions import KGTKException
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.reshape.kgtkidbuilder import KgtkIdBuilder, KgtkIdBuilderOptions
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    input_kgtk_file: Path = KGTKArgumentParser.get_input_file(input_file)
    output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    idbuilder_options: KgtkIdBuilderOptions = KgtkIdBuilderOptions.from_dict(kwargs)
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("--input-file=%s" % str(input_kgtk_file), file=error_file)
        print("--output-file=%s" % str(output_kgtk_file), file=error_file)
        idbuilder_options.show(out=error_file)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    try:

        # First create the KgtkReader.  It provides parameters used by the ID
        # column builder. Next, create the ID column builder, which provides a
        # possibly revised list of column names for the KgtkWriter.  Create
        # the KgtkWriter.  Last, process the data stream.

        # Open the input file.
        kr: KgtkReader =  KgtkReader.open(input_kgtk_file,
                                          error_file=error_file,
                                          options=reader_options,
                                          value_options=value_options,
                                          verbose=verbose,
                                          very_verbose=very_verbose,
        )
        
        # Create the ID builder.
        idb: KgtkIdBuilder = KgtkIdBuilder.new(kr, idbuilder_options)

        # Open the output file.
        ew: KgtkWriter = KgtkWriter.open(idb.column_names,
                                         output_kgtk_file,
                                         mode=KgtkWriter.Mode[kr.mode.name],
                                         require_all_columns=True,
                                         prohibit_extra_columns=True,
                                         fill_missing_columns=False,
                                         gzip_in_parallel=False,
                                         verbose=verbose,
                                         very_verbose=very_verbose)

        # Process the input file, building IDs.
        idb.process(kr, ew)

        # Clean up.
        ew.close()
        kr.close()

        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))




"""
Reorder KGTK file columns (while copying)

TODO: Need KgtkWriterOptions
"""

from argparse import Namespace, SUPPRESS
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Perform calculations on KGTK file columns.',
        'description': 'This command performs calculations on one or more columns in a KGTK file. ' +
        '\nIf no input filename is provided, the default is to read standard input. ' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert rename_columns --help'
    }

# Future:

# Numeric:
ABS_OP: str = "abs" # column
DIV_OP: str = "div" # (column / column) or (column / value)
MOD_OP: str = "mod" # (column mod column) or (column mod value)
POW_OP: str = "pow"
MINUS_OP: str = "minus" # (column - column) or (column - value)
NEGATE_OP: str = "negate"

# String
CENTER_OP: str = "center"
COUNT_OP: str = "count"
ENDSWITH_OP: str = "endswidth"
EXPANDTABS_OP: str = "expandtabs"
FIND_OP: str = "find"
ISALNUM_OP: str = "isalnum"
# ...
LJUST_OP: str = "ljust"
LSTRIP_OP: str = "lstrip"
PARTITION_OP: str = "partition"
REMOVEPREFIX_OP: str = "removeprefix"
REMOVESUFFIX_OP: str = "removesuffix"
RFIND_OP: str = "rfind"
RJUST_OP: str = "rjust"
RPARTITION_OP: str = "rpartition"
RSPLIT_OP: str = "rsplit"
RSTRIP_OP: str = "rstrip"
SPLIT_OP: str = " split"
SPLITLINES_OP: str = "splitlines"
STARTSWITH_OP: str = "startswith"
STRIP_OP: str = "strip"
ZFILL_OP: str = "zfill"

# Implemented:

# Boolean
AND_OP: str = "and" # (boolean, boolean) -> boolean
NAND_OP: str = "nand" # (boolean, boolean) -> boolean
NOR_OP: str = "nor" # (boolean, boolean) -> boolean
NOT_OP: str = "not" # (boolean, ...) -> (boolean, ...)
OR_OP: str = "or" # (boolean, boolean) -> boolean
XOR_OP: str = "xor" # (boolean, boolean) -> boolean

# Numeric
AVERAGE_OP: str = "average"
MAX_OP: str = "max"
MIN_OP: str = "min"
PERCENTAGE_OP: str = "percentage"
SUM_OP: str = "sum" # Sums the columns and the values.
GE_OP: str = "ge" # (column > column) or (column > value) -> boolean
GT_OP: str = "gt" # (column >= column) or (column >= value) -> boolean
LT_OP: str = "lt" # (column < column) or (column < value) -> boolean
LE_OP: str = "le" # (column <= column) or (column <= value) -> boolean
EQ_OP: str = "eq" # (column == column) or (column == value) -> boolean
NE_OP: str = "ne" # (column != column) or (column != value) -> boolean

# String
CAPITALIZE_OP: str = "capitalize"
CASEFOLD_OP: str = "casefold"
JOIN_OP: str = "join"
LOWER_OP: str = "lower"
REPLACE_OP: str = "replace"
SUBSTITUTE_OP: str = "substitute"
SWAPCASE_OP: str = "swapcase"
TITLE_OP: str = "title"
UPPER_OP: str = "upper"

# General
COPY_OP: str = "copy"
IS_OP: str = "is" # (column == column) or (column == value) -> boolean
IS_IN_OP: str = "is_in" # column in values -> boolean
IS_NOT_OP: str = "is_not" # (column != column) or (column != value) -> boolean
SET_OP: str = "set"

# Date/Time
FROMISOFORMAT_OP: str = "fromisoformat"

OPERATIONS: typing.List[str] = [ AND_OP,
                                 AVERAGE_OP,
                                 CAPITALIZE_OP,
                                 CASEFOLD_OP,
                                 COPY_OP,
                                 EQ_OP,
                                 FROMISOFORMAT_OP,
                                 GE_OP,
                                 GT_OP,
                                 IS_OP,
                                 IS_IN_OP,
                                 IS_NOT_OP,
                                 JOIN_OP,
                                 LOWER_OP,
                                 LE_OP,
                                 LT_OP,
                                 MAX_OP,
                                 MIN_OP,
                                 NAND_OP,
                                 NE_OP,
                                 NOR_OP,
                                 NOT_OP,
                                 OR_OP,
                                 PERCENTAGE_OP,
                                 REPLACE_OP,
                                 SET_OP,
                                 SUBSTITUTE_OP,
                                 SUM_OP,
                                 SWAPCASE_OP,
                                 TITLE_OP,
                                 UPPER_OP,
                                 XOR_OP,
                                ]

def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    # import modules locally
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.utils.argparsehelpers import optional_bool
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_input_file()
    parser.add_output_file()

    parser.add_argument(      "--output-format", dest="output_format", help=h("The file format (default=kgtk)"), type=str)

    parser.add_argument('-c', "--columns", dest="column_names_list", nargs='*', metavar="COLUMN_NAME", action='append',
                        help="The list of source column names, optionally containing '..' for column ranges " +
                        "and '...' for column names not explicitly mentioned.")

    parser.add_argument(      "--into", dest="into_column_names_list", nargs='+', metavar="COLUMN_NAME", action='append',
                              help="The name of the column to receive the result of the calculation.",
                              required=True)

    parser.add_argument(      "--do", dest="operation", help="The name of the operation.", required=True,
                              choices=OPERATIONS)

    parser.add_argument(      "--values", dest="values_list", nargs='*', metavar="VALUES", action='append',
                        help="An optional list of values")

    parser.add_argument(      "--with-values", dest="with_values_list", nargs='*', metavar="WITH_VALUES", action='append',
                        help="An optional list of additional values")

    parser.add_argument(      "--limit", dest="limit", type=int,
                              help="A limit count.")

    parser.add_argument(      "--format", dest="format_string", help="The format string for the calculation.")

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def flatten_arg_list(arg: typing.Optional[typing.List[typing.List[str]]])->typing.List[str]:
    result: typing.List[str] = [ ]
    if arg is None:
        return result

    arglist: typing.List[str]
    for arglist in arg:
        value: str
        for value in arglist:
            if value is not None:
                result.append(value)

    return result

def run(input_file: KGTKFiles,
        output_file: KGTKFiles,
        output_format: typing.Optional[str],

        column_names_list: typing.List[typing.List[str]],
        into_column_names_list: typing.List[typing.List[str]],
        operation: str,
        values_list: typing.List[typing.List[str]],
        with_values_list: typing.List[typing.List[str]],
        limit: typing.Optional[int],
        format_string: typing.Optional[str],

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    import datetime as dt
    from pathlib import Path
    import re
    import sys

    from kgtk.exceptions import KGTKException
    from kgtk.kgtkformat import KgtkFormat
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions
    from kgtk.value.kgtkvalue import KgtkValue

    input_kgtk_file: Path = KGTKArgumentParser.get_input_file(input_file)
    output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Flatten the input lists.
    column_names: typing.List[str] = flatten_arg_list(column_names_list)
    into_column_names: typing.List[str] = flatten_arg_list(into_column_names_list)
    values: typing.List[str] = flatten_arg_list(values_list)
    with_values: typing.List[str] = flatten_arg_list(with_values_list)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("--input-file=%s" % str(input_kgtk_file), file=error_file, flush=True)
        print("--output-file=%s" % str(output_kgtk_file), file=error_file, flush=True)
        if output_format is not None:
            print("--output-format=%s" % output_format, file=error_file, flush=True)
        if len(column_names) > 0:
            print("--columns %s" % " ".join(column_names), file=error_file, flush=True)
        if len(into_column_names) > 0:
            print("--into %s" % " ".join(into_column_names), file=error_file, flush=True)
        print("--operation=%s" % str(operation), file=error_file, flush=True)
        if len(values) > 0:
            print("--values %s" % " ".join(values), file=error_file, flush=True)
        if len(with_values) > 0:
            print("--with-values %s" % " ".join(with_values), file=error_file, flush=True)
        if limit is not None:
            print("--limit %d" % limit, file=error_file, flush=True)
        if format_string is not None:
            print("--format=%s" % format_string, file=error_file, flush=True)

        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    try:

        if verbose:
            print("Opening the input file %s" % str(input_kgtk_file), file=error_file, flush=True)
        kr = KgtkReader.open(input_kgtk_file,
                             options=reader_options,
                             value_options = value_options,
                             error_file=error_file,
                             verbose=verbose,
                             very_verbose=very_verbose,
        )

        remaining_names: typing.List[str] = kr.column_names.copy()
        selected_names: typing.List[str] = [ ]
        save_selected_names: typing.Optional[typing.List[str]] = None

        ellipses: str = "..." # All unmentioned columns
        ranger: str = ".." # All columns between two columns.

        idx: int

        saw_ranger: bool = False
        column_name: str
        for column_name in column_names:
            if column_name == ellipses:
                if save_selected_names is not None:
                    raise KGTKException("Elipses may appear only once")

                if saw_ranger:
                    raise KGTKException("Elipses may not appear directly after a range operator ('..').")

                save_selected_names = selected_names
                selected_names = [ ]
                continue

            if column_name == ranger:
                if len(selected_names) == 0:
                    raise KGTKException("The column range operator ('..') may not appear without a preceeding column name.")
                saw_ranger = True
                continue

            if column_name not in kr.column_names:
                raise KGTKException("Unknown column name '%s'." % column_name)
            if column_name not in remaining_names:
                raise KGTKException("Column name '%s' was duplicated in the list." % column_name)

            if saw_ranger:
                saw_ranger = False
                prior_column_name: str = selected_names[-1]
                prior_column_idx: int = kr.column_name_map[prior_column_name]
                column_name_idx: int = kr.column_name_map[column_name]
                start_idx: int
                end_idx: int
                idx_inc: int
                if column_name_idx > prior_column_idx:
                    start_idx = prior_column_idx + 1
                    end_idx = column_name_idx - 1
                    idx_inc = 1
                else:
                    start_idx = prior_column_idx - 1
                    end_idx = column_name_idx + 1
                    idx_inc = -1

                idx = start_idx
                while idx <= end_idx:
                    idx_column_name: str = kr.column_names[idx]
                    if idx_column_name not in remaining_names:
                        raise KGTKException("Column name '%s' (%s .. %s) was duplicated in the list." % (column_name, prior_column_name, column_name))
                   
                    selected_names.append(idx_column_name)
                    remaining_names.remove(idx_column_name)
                    idx += idx_inc

            selected_names.append(column_name)
            remaining_names.remove(column_name)

        if saw_ranger:
            raise KGTKException("The column ranger operator ('..') may not end the list of column names.")

        if len(remaining_names) > 0 and save_selected_names is None:
            if verbose:
                print("Omitting the following columns: %s" % " ".join(remaining_names), file=error_file, flush=True)
        if save_selected_names is not None:
            if len(remaining_names) > 0:
                save_selected_names.extend(remaining_names)
            if len(selected_names) > 0:
                save_selected_names.extend(selected_names)
            selected_names = save_selected_names

        sources: typing.List[int] = [ ]
        name: str
        for name in selected_names:
            sources.append(kr.column_name_map[name])

        new_column_count: int = 0
        into_column_idxs: typing.List[int] = [ ]
        into_column_idx: int
        output_column_names: typing.List[str] = kr.column_names.copy()
        into_column_name: str
        for idx, into_column_name in enumerate(into_column_names):
            if into_column_name in kr.column_name_map:
                into_column_idx = kr.column_name_map[into_column_name]
                into_column_idxs.append(into_column_idx)
                if verbose:
                    print("Putting result %d of the calculation into old column %d (%s)." % (idx + 1, into_column_idx, into_column_name), file=error_file, flush=True)
            else:
                new_column_count += 1
                into_column_idx = len(output_column_names)
                into_column_idxs.append(into_column_idx)
                output_column_names.append(into_column_name)
                if verbose:
                    print("Putting result %d of the calculation into new column %d (%s)." % (idx + 1, into_column_idx, into_column_name), file=error_file, flush=True)

        if verbose:
            print("Opening the output file %s" % str(output_kgtk_file), file=error_file, flush=True)
        kw: KgtkWriter = KgtkWriter.open(output_column_names,
                                         output_kgtk_file,
                                         require_all_columns=True,
                                         prohibit_extra_columns=True,
                                         fill_missing_columns=False,
                                         gzip_in_parallel=False,
                                         mode=KgtkWriter.Mode[kr.mode.name],
                                         output_format=output_format,
                                         verbose=verbose,
                                         very_verbose=very_verbose,
        )

        if limit is None:
            limit = 0

        substitute_re: typing.Optional[typing.Pattern] = None

        if operation == AND_OP:
            if len(sources) == 0:
                raise KGTKException("And needs at least one source, got %d" % len(sources))
            if len(into_column_idxs) != 1:
                raise KGTKException("And needs 1 destination column, got %d" % len(into_column_idxs))

        elif operation == AVERAGE_OP:
            if len(sources) == 0:
                raise KGTKException("Average needs at least one source, got %d" % len(sources))
            if len(into_column_idxs) != 1:
                raise KGTKException("Average needs 1 destination column, got %d" % len(into_column_idxs))

        elif operation == CAPITALIZE_OP:
            if len(sources) == 0:
                raise KGTKException("Capitalize needs at least one source, got %d" % len(sources))
            if len(sources) != len(into_column_idxs):
                raise KGTKException("Capitalize needs the same number of input columns and into columns, got %d and %d" % (len(sources), len(into_column_idxs)))

        elif operation == CASEFOLD_OP:
            if len(sources) == 0:
                raise KGTKException("Casefold needs at least one source, got %d" % len(sources))
            if len(sources) != len(into_column_idxs):
                raise KGTKException("Casefold needs the same number of input columns and into columns, got %d and %d" % (len(sources), len(into_column_idxs)))

        elif operation == COPY_OP:
            if len(sources) == 0:
                raise KGTKException("Copy needs at least one source, got %d" % len(sources))
            if len(selected_names) != len(into_column_idxs):
                raise KGTKException("Copy needs the same number of input columns and into columns, got %d and %d" % (len(selected_names), len(into_column_idxs)))

        elif operation == EQ_OP:
            if (len(sources) == 2 and len(values) == 0) or (len(sources) == 1 and len(values) == 1):
                raise KGTKException("Eq needs two sources or one source and one value, got %d sources and %d values" % (len(sources), len(values)))
            if len(into_column_idxs) != 1:
                raise KGTKException("Eq needs 1 destination columns, got %d" % len(into_column_idxs))

        elif operation == FROMISOFORMAT_OP:
            if len(sources) != 1:
                raise KGTKException("Fromisoformat needs one source, got %d" % len(sources))
            if len(values) != len(into_column_idxs):
                raise KGTKException("Fromisoformat needs the same number of values and into columns, got %d and %d" % (len(values), len(into_column_idxs)))

        elif operation == GE_OP:
            if (len(sources) == 2 and len(values) == 0) or (len(sources) == 1 and len(values) == 1):
                raise KGTKException("Ge needs two sources or one source and one value, got %d sources and %d values" % (len(sources), len(values)))
            if len(into_column_idxs) != 1:
                raise KGTKException("Ge needs 1 destination columns, got %d" % len(into_column_idxs))

        elif operation == GT_OP:
            if (len(sources) == 2 and len(values) == 0) or (len(sources) == 1 and len(values) == 1):
                raise KGTKException("Gt needs two sources or one source and one value, got %d sources and %d values" % (len(sources), len(values)))
            if len(into_column_idxs) != 1:
                raise KGTKException("Gt needs 1 destination columns, got %d" % len(into_column_idxs))

        elif operation == IS_OP:
            if (len(sources) == 2 and len(values) == 0) or (len(sources) == 1 and len(values) == 1):
                raise KGTKException("Is needs two sources or one source and one value, got %d sources and %d values" % (len(sources), len(values)))
            if len(into_column_idxs) != 1:
                raise KGTKException("Is needs 1 destination columns, got %d" % len(into_column_idxs))

        elif operation == IS_IN_OP:
            if len(sources) != 1:
                raise KGTKException("Is in needs one source, got %d" % len(sources))
            if len(values) == 0:
                raise KGTKException("Is in needs at least one value, got %d" % len(values))
            if len(into_column_idxs) != 1:
                raise KGTKException("Is in needs 1 destination columns, got %d" % len(into_column_idxs))

        elif operation == IS_NOT_OP:
            if (len(sources) == 2 and len(values) == 0) or (len(sources) == 1 and len(values) == 1):
                raise KGTKException("Is not needs two sources or one source and one value, got %d sources and %d values" % (len(sources), len(values)))
            if len(into_column_idxs) != 1:
                raise KGTKException("Is not needs 1 destination columns, got %d" % len(into_column_idxs))

        elif operation == JOIN_OP:
            if len(sources) == 0:
                raise KGTKException("Join needs at least one source, got %d" % len(sources))
            if len(into_column_idxs) != 1:
                raise KGTKException("Join needs 1 destination columns, got %d" % len(into_column_idxs))
            if len(values) != 1:
                raise KGTKException("Join needs 1 value, got %d" % len(values))

        elif operation == LE_OP:
            if (len(sources) == 2 and len(values) == 0) or (len(sources) == 1 and len(values) == 1):
                raise KGTKException("Le needs two sources or one source and one value, got %d sources and %d values" % (len(sources), len(values)))
            if len(into_column_idxs) != 1:
                raise KGTKException("Le needs 1 destination columns, got %d" % len(into_column_idxs))

        elif operation == LT_OP:
            if (len(sources) == 2 and len(values) == 0) or (len(sources) == 1 and len(values) == 1):
                raise KGTKException("Lt needs two sources or one source and one value, got %d sources and %d values" % (len(sources), len(values)))
            if len(into_column_idxs) != 1:
                raise KGTKException("Lt needs 1 destination columns, got %d" % len(into_column_idxs))

        elif operation == LOWER_OP:
            if len(sources) == 0:
                raise KGTKException("Lower needs at least one source, got %d" % len(sources))
            if len(sources) != len(into_column_idxs):
                raise KGTKException("Lower needs the same number of input columns and into columns, got %d and %d" % (len(sources), len(into_column_idxs)))

        elif operation == MAX_OP:
            if len(sources) == 0:
                raise KGTKException("Max needs at least one source, got %d" % len(sources))
            if len(into_column_idxs) != 1:
                raise KGTKException("Max needs 1 destination columns, got %d" % len(into_column_idxs))

        elif operation == MIN_OP:
            if len(sources) == 0:
                raise KGTKException("Min needs at least one source, got %d" % len(sources))
            if len(into_column_idxs) != 1:
                raise KGTKException("Min needs 1 destination columns, got %d" % len(into_column_idxs))

        elif operation == NE_OP:
            if (len(sources) == 2 and len(values) == 0) or (len(sources) == 1 and len(values) == 1):
                raise KGTKException("Ne needs two sources or one source and one value, got %d sources and %d values" % (len(sources), len(values)))
            if len(into_column_idxs) != 1:
                raise KGTKException("Ne needs 1 destination columns, got %d" % len(into_column_idxs))

        elif operation == NOR_OP:
            if len(sources) == 0:
                raise KGTKException("Nor needs at least one source, got %d" % len(sources))
            if len(into_column_idxs) != 1:
                raise KGTKException("Nor needs 1 destination column, got %d" % len(into_column_idxs))

        elif operation == NOT_OP:
            if len(sources) == 0:
                raise KGTKException("Not needs at least one source, got %d" % len(sources))
            if len(into_column_idxs) != len(sources):
                raise KGTKException("Nand needs the same number of input columns and into colums, got %d and %d" % (len(sources), len(into_column_idxs)))

        elif operation == OR_OP:
            if len(sources) == 0:
                raise KGTKException("Or needs at least one source, got %d" % len(sources))
            if len(into_column_idxs) != 1:
                raise KGTKException("Or needs 1 destination column, got %d" % len(into_column_idxs))

        elif operation == PERCENTAGE_OP:
            if len(into_column_idxs) != 1:
                raise KGTKException("Percent needs 1 destination columns, got %d" % len(into_column_idxs))
            if len(selected_names) != 2:
                raise KGTKException("Percent needs 2 input columns, got %d" % len(selected_names))

        elif operation == REPLACE_OP:
            if len(into_column_idxs) != 1:
                raise KGTKException("Replace needs 1 destination column, got %d" % len(into_column_idxs))
            if len(selected_names) != 1:
                raise KGTKException("Replace needs 1 input column, got %d" % len(selected_names))
            if len(values) != 1:
                raise KGTKException("Replace needs one value, got %d" % len(values))
            if len(with_values) != 1:
                raise KGTKException("Replace needs one with-value, got %d" % len(with_values))

        elif operation == SET_OP:
            if len(sources) != 0:
                raise KGTKException("Set needs no sources, got %d" % len(sources))
            if len(into_column_idxs) == 0:
                raise KGTKException("Set needs at least one destination column, got %d" % len(into_column_idxs))
            if len(values) == 0:
                raise KGTKException("Set needs at least one value, got %d" % len(values))
            if len(into_column_idxs) != len(values):
                raise KGTKException("Set needs the same number of destination columns and values, got %d and %d" % (len(into_column_idxs), len(values)))

        elif operation == SUBSTITUTE_OP:
            if len(into_column_idxs) != 1:
                raise KGTKException("Substitute needs 1 destination column, got %d" % len(into_column_idxs))
            if len(selected_names) != 1:
                raise KGTKException("Substitute needs 1 input column, got %d" % len(selected_names))
            if len(values) != 1:
                raise KGTKException("Substitute needs one value, got %d" % len(values))
            if len(with_values) != 1:
                raise KGTKException("Substitute needs one with-value, got %d" % len(with_values))
            substitute_re = re.compile(values[0])

        elif operation == SUM_OP:
            if len(sources) == 0:
                raise KGTKException("Sum needs at least one source, got %d" % len(sources))
            if len(into_column_idxs) != 1:
                raise KGTKException("Sum needs 1 destination columns, got %d" % len(into_column_idxs))

        elif operation == SWAPCASE_OP:
            if len(sources) == 0:
                raise KGTKException("Swapcase needs at least one source, got %d" % len(sources))
            if len(sources) != len(into_column_idxs):
                raise KGTKException("Swapcase needs the same number of input columns and into columns, got %d and %d" % (len(sources), len(into_column_idxs)))

        elif operation == TITLE_OP:
            if len(sources) == 0:
                raise KGTKException("Title needs at least one source, got %d" % len(sources))
            if len(sources) != len(into_column_idxs):
                raise KGTKException("Title needs the same number of input columns and into columns, got %d and %d" % (len(sources), len(into_column_idxs)))

        elif operation == UPPER_OP:
            if len(sources) == 0:
                raise KGTKException("Upper needs at least one source, got %d" % len(sources))
            if len(sources) != len(into_column_idxs):
                raise KGTKException("Upper needs the same number of input columns and into columns, got %d and %d" % (len(sources), len(into_column_idxs)))

        elif operation == XOR_OP:
            if len(sources) == 0:
                raise KGTKException("Xor needs at least one source, got %d" % len(sources))
            if len(into_column_idxs) != 1:
                raise KGTKException("Xor needs 1 destination column, got %d" % len(into_column_idxs))


        fs: str = format_string if format_string is not None else "%5.2f"
        item: str
        item2: str
        kv: KgtkValue
        bresult: bool

        into_column_idx = into_column_idxs[0] # for convenience

        input_data_lines: int = 0
        row: typing.List[str]
        for row in kr:
            input_data_lines += 1

            output_row: typing.List[str] = row.copy()
            for idx in range(new_column_count):
                output_row.append("") # Easiest way to add a new column.

            if operation == AND_OP:
                bresult = True
                for idx in sources:
                    kv = KgtkValue(row[idx])
                    if kv.is_boolean():
                        bresult = bresult and kv.is_true()

                output_row[into_column_idx] = KgtkValue.to_boolean(bresult)

            elif operation == AVERAGE_OP:
                atotal: float = 0
                acount: int = 0
                for idx in sources:
                    item = row[idx]
                    if len(item) > 0:
                        atotal += float(item)
                        acount += 1
                output_row[into_column_idx] = (fs % (atotal / float(acount))) if acount > 0 else ""                

            elif operation == CAPITALIZE_OP:
                for idx in range(len(sources)):
                    output_row[into_column_idxs[idx]] = row[sources[idx]].capitalize()

            elif operation == CASEFOLD_OP:
                for idx in range(len(sources)):
                    output_row[into_column_idxs[idx]] = row[sources[idx]].casefold()

            elif operation == COPY_OP:
                for idx in range(len(sources)):
                    output_row[into_column_idxs[idx]] = row[sources[idx]]

            elif operation == EQ_OP:
                if len(sources) == 1:
                    if len(row[sources[0]]) > 0 and len(row[sources[1]]) > 0:
                        output_row[into_column_idx] = KgtkValue.to_boolean(float(row[sources[0]]) == float(row[sources[1]]))
                    else:
                        output_row[into_column_idx] = ""
                else:
                    if len(row[sources[0]]) > 0 and len(values[0]) > 0:
                        output_row[into_column_idx] = KgtkValue.to_boolean(float(row[sources[0]]) == float(values[0]))
                    else:
                        output_row[into_column_idx] = ""

            elif operation == FROMISOFORMAT_OP:
                dtval: str = row[sources[0]]
                if dtval.startswith(KgtkFormat.DATE_AND_TIMES_SIGIL):
                    kgtkdatestr: str = row[sources[0]][1:] # Strip the leading ^
                    isodatestr: str
                    precisionstr: str
                    if "/" in kgtkdatestr:
                        isodatestr, precisionstr = kgtkdatestr.split("/")
                    else:
                        isodatestr = kgtkdatestr
                        precisionstr = ""
                    if isodatestr.endswith("Z"):
                        isodatestr = isodatestr[:-1]

                    into_idx: int
                    value_name: str
                    try:
                        dtvar: dt.datetime = dt.datetime.fromisoformat(isodatestr)
                        for idx in range(len(values)):
                            value_name = values[idx]
                            into_idx = into_column_idxs[idx]
                            
                            if value_name == "year":
                                output_row[into_idx] = str(dtvar.year)

                            elif value_name == "month":
                                output_row[into_idx] = str(dtvar.month)
                    
                            elif value_name == "day":
                                output_row[into_idx] = str(dtvar.day)

                            elif value_name == "hour":
                                output_row[into_idx] = str(dtvar.hour)
                    
                            elif value_name == "minute":
                                output_row[into_idx] = str(dtvar.minute)
                    
                            elif value_name == "second":
                                output_row[into_idx] = str(dtvar.second)
                    
                            elif value_name == "microsecond":
                                output_row[into_idx] = str(dtvar.microsecond)

                            elif value_name == "error":
                                output_row[into_idx] = ""

                            else:
                                raise KGTKException("Unknown date component %s" % repr(value_name))

                    except ValueError as e:
                        print("Error parsing %s in [%s]: %s" % (repr(isodatestr), "|".join([repr(x) for x in row]), str(e)),
                              file=error_file, flush=True)

                        for idx in range(len(values)):
                            value_name = values[idx]
                            into_idx = into_column_idxs[idx]
                            if value_name == "error":
                                output_row[into_idx] = str(e)
                            else:
                                output_row[into_idx] = ""

                else:
                    # Not a date/time value, clear the result columns.
                    for idx in range(len(values)):
                        output_row[into_column_idxs[idx]] = ""
                    
            elif operation == GE_OP:
                if len(sources) == 1:
                    if len(row[sources[0]]) > 0 and len(row[sources[1]]) > 0:
                        output_row[into_column_idx] = KgtkValue.to_boolean(float(row[sources[0]]) >= float(row[sources[1]]))
                    else:
                        output_row[into_column_idx] = ""
                else:
                    if len(row[sources[0]]) > 0 and len(values[0]) > 0:
                        output_row[into_column_idx] = KgtkValue.to_boolean(float(row[sources[0]]) >= float(values[0]))
                    else:
                        output_row[into_column_idx] = ""

            elif operation == GT_OP:
                if len(sources) == 1:
                    if len(row[sources[0]]) > 0 and len(row[sources[1]]) > 0:
                        output_row[into_column_idx] = KgtkValue.to_boolean(float(row[sources[0]]) > float(row[sources[1]]))
                    else:
                        output_row[into_column_idx] = ""
                else:
                    if len(row[sources[0]]) > 0 and len(values[0]) > 0:
                        output_row[into_column_idx] = KgtkValue.to_boolean(float(row[sources[0]]) > float(values[0]))
                    else:
                        output_row[into_column_idx] = ""

            elif operation == IS_OP:
                if len(sources) == 1:
                    output_row[into_column_idx] = KgtkValue.to_boolean(row[sources[0]] == row[sources[1]])
                else:
                    output_row[into_column_idx] = KgtkValue.to_boolean(row[sources[0]] == values[0])

            elif operation == IS_IN_OP:
                bresult = False
                item = row[sources[0]]
                for item2 in values:
                    if item == item2:
                        bresult = True
                        break
                output_row[into_column_idx] = KgtkValue.to_boolean(bresult)

            elif operation == IS_NOT_OP:
                if len(sources) == 1:
                    output_row[into_column_idx] = KgtkValue.to_boolean(row[sources[0]] != row[sources[1]])
                else:
                    output_row[into_column_idx] = KgtkValue.to_boolean(row[sources[0]] != values[0])

            elif operation == JOIN_OP:
                output_row[into_column_idx] = values[0].join((row[sources[idx]] for idx in range(len(sources))))

            elif operation == LE_OP:
                if len(sources) == 1:
                    if len(row[sources[0]]) > 0 and len(row[sources[1]]) > 0:
                        output_row[into_column_idx] = KgtkValue.to_boolean(float(row[sources[0]]) <= float(row[sources[1]]))
                    else:
                        output_row[into_column_idx] = ""
                else:
                    if len(row[sources[0]]) > 0 and len(values[0]) > 0:
                        output_row[into_column_idx] = KgtkValue.to_boolean(float(row[sources[0]]) <= float(values[0]))
                    else:
                        output_row[into_column_idx] = ""

            elif operation == LT_OP:
                if len(sources) == 1:
                    if len(row[sources[0]]) > 0 and len(row[sources[1]]) > 0:
                        output_row[into_column_idx] = KgtkValue.to_boolean(float(row[sources[0]]) < float(row[sources[1]]))
                    else:
                        output_row[into_column_idx] = ""
                else:
                    if len(row[sources[0]]) > 0 and len(values[0]) > 0:
                        output_row[into_column_idx] = KgtkValue.to_boolean(float(row[sources[0]]) < float(values[0]))
                    else:
                        output_row[into_column_idx] = ""

            elif operation == LOWER_OP:
                for idx in range(len(sources)):
                    output_row[into_column_idxs[idx]] = row[sources[idx]].lower()

            elif operation == MAX_OP:
                max_result: typing.Optional[float] = None
                for idx in sources:
                    item = row[idx]
                    if len(item) > 0:
                        max_value: float = float(item)
                        if max_result is None or max_value > max_result:
                            max_result = max_value
                output_row[into_column_idx] = (fs % max_result) if max_result is not None else ""

            elif operation == MIN_OP:
                min_result: typing.Optional[float] = None
                for idx in sources:
                    item = row[idx]
                    if len(item) > 0:
                        min_value: float = float(item)
                        if min_result is None or min_value < min_result:
                            min_result = min_value
                output_row[into_column_idx] = (fs % min_result) if min_result is not None else ""

            elif operation == NAND_OP:
                bresult = True
                for idx in sources:
                    kv = KgtkValue(row[idx])
                    if kv.is_boolean():
                        bresult = bresult and kv.is_true()

                output_row[into_column_idx] = KgtkValue.to_boolean(not bresult)

            elif operation == NE_OP:
                if len(sources) == 1:
                    if len(row[sources[0]]) > 0 and len(row[sources[1]]) > 0:
                        output_row[into_column_idx] = KgtkValue.to_boolean(float(row[sources[0]]) != float(row[sources[1]]))
                    else:
                        output_row[into_column_idx] = ""
                else:
                    if len(row[sources[0]]) > 0 and len(values[0]) > 0:
                        output_row[into_column_idx] = KgtkValue.to_boolean(float(row[sources[0]]) != float(values[0]))
                    else:
                        output_row[into_column_idx] = ""

            elif operation == NOR_OP:
                bresult = False
                for idx in sources:
                    kv = KgtkValue(row[idx])
                    if kv.is_boolean():
                        bresult = bresult or kv.is_true()

                output_row[into_column_idx] = KgtkValue.to_boolean(not bresult)

            elif operation == NOT_OP:
                for idx in sources:
                    kv = KgtkValue(row[idx])
                    if kv.is_boolean():
                        output_row[into_column_idxs[idx]] = KgtkValue.to_boolean(not kv.is_true())
                    else:
                        output_row[into_column_idxs[idx]] = ""

            elif operation == OR_OP:
                bresult = False
                for idx in sources:
                    kv = KgtkValue(row[idx])
                    if kv.is_boolean():
                        bresult = bresult or kv.is_true()

                output_row[into_column_idx] = KgtkValue.to_boolean(bresult)

            elif operation == PERCENTAGE_OP:
                output_row[into_column_idx] = fs % (float(row[sources[0]]) * 100 / float(row[sources[1]]))

            elif operation == REPLACE_OP:
                if limit == 0:
                    output_row[into_column_idx] = row[sources[0]].replace(values[0], with_values[0])
                else:
                    output_row[into_column_idx] = row[sources[0]].replace(values[0], with_values[0], limit)

            elif operation == SET_OP:
                for idx in range(len(values)):
                    output_row[into_column_idxs[idx]] = values[idx]

            elif operation == SUBSTITUTE_OP and substitute_re is not None:
                output_row[into_column_idx] = substitute_re.sub(with_values[0], row[sources[0]], count=limit)

            elif operation == SUM_OP:
                total: float = 0
                for idx in sources:
                    item = row[idx]
                    if len(item) > 0:
                        total += float(item)
                for item in values:
                    if len(item) > 0:
                        total += float(item)
                output_row[into_column_idx] = fs % total
                
            elif operation == SWAPCASE_OP:
                for idx in range(len(sources)):
                    output_row[into_column_idxs[idx]] = row[sources[idx]].swapcase()

            elif operation == TITLE_OP:
                for idx in range(len(sources)):
                    output_row[into_column_idxs[idx]] = row[sources[idx]].title()

            elif operation == UPPER_OP:
                for idx in range(len(sources)):
                    output_row[into_column_idxs[idx]] = row[sources[idx]].upper()

            elif operation == XOR_OP:
                bresult = False
                for idx in sources:
                    kv = KgtkValue(row[idx])
                    if kv.is_boolean():
                        bresult = bresult != kv.is_true()

                output_row[into_column_idx] = KgtkValue.to_boolean(bresult)

            kw.write(output_row)

        # Flush the output file so far:
        kw.flush()

        if verbose:
            print("Read %d data lines from file %s" % (input_data_lines, input_kgtk_file), file=error_file, flush=True)

        kw.close()

        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))




"""This KGTK command runs the Posix sort command to sort KGTK files.  In normal operation,
the system sort program is used to sort the data, which is expected to provide
good performance when properly configured.  The system sort program is run in
a background data processing pipeline that runs in parallel with the Python
control process.  The Python control process and the data processing pipeline
interact in a clever way to allow the Python code to use KgtkReader to process
the input file's header line without sending the rest of the data through
Python's expensive I/O path.

1) The data processing pipeline reads stdin or a named file.
   The named file is fed to the data processing pipeline by `cat`,
   avoiding having it read by Python.

2) The header line is stripped out of the input stream by a
   shell `read` command.

3) The header line is then copied to the output stream using a shell
   'printf' command.

4) A copy of the header line is sent via a pipe to the Python control
   process.

5) The data processing pipeline then waits to read sort options
   from a second pipe.

6) The Python control process reads the header line from the dirst pipe

7) The Python control process feeds the header line to KgtkReader.

8) The Python control process builds the sort key options using
   column indexes obtained from KgtkReader.

9) The sort key options are sent from Python to the data processing pipeline
   via the second pipe.

10) The data processing pipeline receives the sort command options via
   the shell `read` command, and passes them to the `sort` program.

11) The sort command reads the rest of the input stream,
    sorts it, and writes the sorted data to the output stream.

An alternate mode of operation is available in which pure Python code is used
to sort the input data.  This mode is more portable but slower and mor memory
intensive than using the system sort program to sort the input KGTK file.

"""
from argparse import Namespace, SUPPRESS
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles
from kgtk.exceptions import KGTKException

def parser():
    return {
        'help': 'Sort file based on one or more columns',
        'aliases': ['sort2']
    }

def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.utils.argparsehelpers import optional_bool
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_input_file(positional=True, metavar="INPUT",
                          who="Input file to sort.")
    parser.add_output_file(options=['-o', '--out', '--output-file'],
                           who='Output file to write to.')

    parser.add_argument('-c', '--column', '--columns', action='store', dest='columns', nargs='*',
                        help="space and/or comma-separated list of column names to sort on (the key columns). " +
                        "(defaults to id for node files, " +
                        "(node1, label, node2) for edge files without ID, (id, node1, label, node2) for edge files with ID)")

    parser.add_argument(      '--locale', dest='locale', type=str, default='C',
                              help="LC_ALL locale controls the sorting order. (default=%(default)s)")

    parser.add_argument('-r', '--reverse', dest='reverse_sort', metavar="True|False",
                        help="When True, generate output in reverse (descending) sort order.  All key columns are sorted in reverse order. (default=%(default)s)",
                        type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      '--reverse-columns', action='store', dest='reverse_columns', nargs='*',
                        help="List specific key columns for reverse (descending) sorting. Overidden by --reverse. (default=none)")
                       
    parser.add_argument(      '--numeric', dest='numeric_sort', metavar="True|False",
                        help="When True, generate output in numeric sort order. All key columns are sorted in numeric order. (default=%(default)s)",
                        type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      '--numeric-columns', action='store', dest='numeric_columns', nargs='*',
                        help="List specific key columns for numeric sorting. Overridden by --numeric. (default=none)")

    parser.add_argument(      '--pure-python', dest='pure_python', metavar="True|False",
                        help="When True, sort in-memory with Python code. (default=%(default)s)",
                        type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument('-X', '--extra', default='', action='store', dest='extra',
                        help="extra options to supply to the sort program. (default=None)")

    parser.add_argument(      '--bash-command', dest='bash_command', type=str, default="bash",
                        help=h("The bash command or its substitute. (default=%(default)s)"))

    parser.add_argument(      '--bzip2-command', dest='bzip2_command', type=str, default="bzip2",
                        help=h("The bzip2 command or its substitute. (default=%(default)s)"))

    parser.add_argument(      '--gzip-command', dest='gzip_command', type=str, default="gzip",
                        help=h("The gzip command or its substitute. (default=%(default)s)"))

    parser.add_argument(      '--pgrep-command', dest='pgrep_command', type=str, default="pgrep",
                        help=h("The pgrep command or its substitute. (default=%(default)s)"))

    parser.add_argument(      '--sort-command', dest='sort_command', type=str, default="sort",
                        help=h("The sort command or its substitute. (default=%(default)s)"))

    parser.add_argument(      '--xz-command', dest='xz_command', type=str, default="xz",
                        help=h("The xz command or its substitute. (default=%(default)s)"))

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def custom_progress()->bool:
    return True # We want to start a custom progress monitor.

header_read_fd : int = -1
header_write_fd: int = -1
sortopt_read_fd : int = -1
sortopt_write_fd: int = -1
cat_proc = None
cmd_proc = None
def cleanup():
    # Clean up the file descriptors and processes, just in case.
    import os
    if header_read_fd >= 0:
        try:
            os.close(header_read_fd)
        except os.error:
            pass

    if header_write_fd >= 0:
        try:
            os.close(header_write_fd)
        except os.error:
            pass
            
    if sortopt_read_fd >= 0:
        try:
            os.close(sortopt_read_fd)
        except os.error:
            pass
            
    if sortopt_write_fd >= 0:
        try:
            os.close(sortopt_write_fd)
        except os.error:
            pass
            
    global cat_proc
    if cat_proc is not None:
        try:
            cat_proc.kill_group()
        except os.error:
            pass
        cat_proc = None

    global cmd_proc
    if cmd_proc is not None:
        try:
            cmd_proc.kill_group()
        except os.error:
            pass
        cat_proc = None

def keyboard_interrupt():
    cleanup()

def run(input_file: KGTKFiles,
        output_file: KGTKFiles,
        columns: typing.Optional[typing.List[str]] = None,
        locale: str = "C",
        reverse_sort: bool = False,
        reverse_columns: typing.Optional[typing.List[str]] = None,
        numeric_sort: bool = False,
        numeric_columns: typing.Optional[typing.List[str]] = None,
        pure_python: bool = False,
        extra: typing.Optional[str] = None,

        bash_command: str = "bash",
        bzip2_command: str = "bzip2",
        gzip_command: str = "gzip",
        pgrep_command: str = "pgrep",
        sort_command: str = "sort",
        xz_command: str = "xz",

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.       
)->int:
    from io import StringIO
    import os
    from pathlib import Path
    import sh # type: ignore
    import sys
    import typing

    from kgtk.cli_entry import progress_startup
    from kgtk.kgtkformat import KgtkFormat
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    input_path: Path = KGTKArgumentParser.get_input_file(input_file)
    output_path: Path = KGTKArgumentParser.get_output_file(output_file)

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    def python_sort():
        if numeric_columns is not None and len(numeric_columns) > 0:
            raise KGTKException('Error: the pure Python sorter does not currently support numeric column sorts.')

        if reverse_columns is not None and len(reverse_columns) > 0:
            raise KGTKException('Error: the pure Python sorter does not currently support reverse column sorts.')

        if verbose:
            print("Opening the input file: %s" % str(input_path), file=error_file, flush=True)
        kr: KgtkReader = KgtkReader.open(input_path,
                                         options=reader_options,
                                         value_options = value_options,
                                         error_file=error_file,
                                         verbose=verbose,
                                         very_verbose=very_verbose,
        )

        sort_idx: int
        key_idxs: typing.List[int] = [ ]
        if columns is not None and len(columns) > 0:
            # Process the list of column names, including splitting
            # comma-separated lists of column names.
            column_name: str
            for column_name in columns:
                column_name_2: str
                for column_name_2 in column_name.split(","):
                    column_name_2 = column_name_2.strip()
                    if len(column_name_2) == 0:
                        continue
                    if column_name_2.isdigit():
                        sort_idx = int(column_name_2)
                        if sort_idx > len(kr.column_names):
                            kr.close()
                            cleanup()
                            raise KGTKException("Invalid column number %d (max %d)." % (sort_idx, len(kr.column_names)))
                        key_idxs.append(sort_idx - 1)
                    else:
                        if column_name_2 not in kr.column_names:
                            kr.close()
                            cleanup()
                            raise KGTKException("Unknown column_name %s" % column_name_2)
                        key_idxs.append(kr.column_name_map[column_name_2])
        else:
            if kr.is_node_file:
                key_idxs.append(kr.id_column_idx)

            elif kr.is_edge_file:
                if kr.id_column_idx >= 0:
                    key_idxs.append(kr.id_column_idx)

                key_idxs.append(kr.node1_column_idx)
                key_idxs.append(kr.label_column_idx)
                key_idxs.append(kr.node2_column_idx)
            else:
                cleanup()
                raise KGTKException("Unknown KGTK file mode, please specify the sorting columns.")

        if verbose:
            print("sorting keys: %s" % " ".join([str(x) for x in key_idxs]), file=error_file, flush=True)

        if numeric_sort and len(key_idxs) > 1:
            raise KGTKException('Error: the pure Python sorter does not currently support numeric sorts on multiple columns.')

        lines: typing.MutableMapping[typing.Union[str, float], typing.List[typing.List[str]]] = dict()

        progress_startup()
        key: typing.Union[str, float]
        row: typing.List[str]
        for row in kr:
            key = KgtkFormat.KEY_FIELD_SEPARATOR.join(row[idx] for idx in key_idxs)
            if numeric_sort:
                key = float(key)
            if key in lines:
                # There are multiple rows with the same key.  Make this a stable sort.
                lines[key].append(row)
            else:
                lines[key] = [ row ]
        if verbose:
            print("\nRead %d data lines." % len(lines), file=error_file, flush=True)

        kw = KgtkWriter.open(kr.column_names,
                             output_path,
                             mode=KgtkWriter.Mode[kr.mode.name],
                             verbose=verbose,
                             very_verbose=very_verbose)
        
        for key in sorted(lines.keys(), reverse=reverse_sort):
            for row in lines[key]:
                kw.write(row)

        kw.close()
        kr.close()

    if pure_python:
        return python_sort()

    try:
        global header_read_fd
        global header_write_fd
        header_read_fd, header_write_fd = os.pipe()
        os.set_inheritable(header_write_fd, True)
        if verbose:
            print("header pipe: read_fd=%d write_fd=%d" % (header_read_fd, header_write_fd), file=error_file, flush=True)
        
        global sortopt_read_fd
        global sortopt_write_fd
        sortopt_read_fd, sortopt_write_fd = os.pipe()
        os.set_inheritable(sortopt_read_fd, True)
        if verbose:
            print("sort options pipe: read_fd=%d write_fd=%d" % (sortopt_read_fd, sortopt_write_fd), file=error_file, flush=True)

        locale_envar: str = "LC_ALL=%s" % locale if len(locale) > 0 else ""

        # Note: "read -u n", used below, is not supported by some shells.
        # bash and zsh support it.
        # ash, csh, dash, and tcsh do not.
        # The original standard Bourne shell, sh, does not.
        # ksh might do it, if the FD number is a single digit.
        cmd: str = "".join((
            "{ IFS= read -r header ; ", # Read the header line
            " { printf \"%s\\n\" \"$header\" >&" +  str(header_write_fd) + " ; } ; ", # Send the header to Python
            " printf \"%s\\n\" \"$header\" ; ", # Send the header to standard output (which may be redirected to a file, below).
            " IFS= read -u " + str(sortopt_read_fd) + " -r options ; ", # Read the sort command options from Python.
            " %s %s -t '\t' $options ; } " % (locale_envar, sort_command), # Sort the remaining input lines using the options read from Python.
        ))
        if str(output_path) != "-":
            # Do we want to compress the output?
            output_suffix: str = output_path.suffix.lower()
            if output_suffix in [".gz", ".z"]:
                if verbose:
                    print("gzip output file: %s" % repr(str(output_path)), file=error_file, flush=True)
                cmd += " | " + gzip_command + " -"

            elif output_suffix in [".bz2", ".bz"]:
                if verbose:
                    print("bzip2 output file: %s" % repr(str(output_path)), file=error_file, flush=True)
                cmd += " | " + bzip2_command + " -z"

            elif output_suffix in [".xz", ".lzma"]:
                if verbose:
                    print("xz output file: %s" % repr(str(output_path)), file=error_file, flush=True)
                cmd += " | " + xz_command + " -z -"

            # Feed the sorted output to the named file.  Otherwise, the sorted
            # output goes to standard output without passing through Python.
            cmd += " > " + repr(str(output_path))

        if verbose:
            print("sort command: %s" % cmd, file=error_file, flush=True)

        global cat_proc
        cat_proc = None
        global cmd_proc
        cmd_proc = None

        def cat_done(cmd, success, exit_code):
            # When the cat command finishes, monitor the progress of the sort command.
            if verbose:
                print("\nDone reading the input file", file=error_file, flush=True)
            if cmd_proc is None:
                return

            # Locate the sort command using pgrep
            buf = StringIO()
            try:
                sh_pgrep = sh.Command(pgrep_command)
                sh_pgrep("-g", cmd_proc.pgid, "--newest", sort_command, _out=buf)
                pgrep_output = buf.getvalue()
                if len(pgrep_output) == 0:
                    if verbose:
                        print("Unable to locate the sort command.", file=error_file, flush=True)
                    return
                sort_pid = int(pgrep_output)
            except Exception as e:
                if verbose:
                    print("Exception looking for sort command: %s" % str(e), file=error_file, flush=True)
                return
            finally:
                buf.close()

            if verbose:
                print("Monitoring the sort command (pid=%d)" % sort_pid, file=error_file, flush=True)
            progress_startup(pid=sort_pid)

        if str(input_path) == "-":
            # Read from standard input.
            #
            # Sh version 1.13 or greater is required for _pass_fds.
            sh_bash = sh.Command(bash_command)
            cmd_proc = sh_bash("-c", cmd, _in=sys.stdin, _out=sys.stdout, _err=sys.stderr,
                               _bg=True, _bg_exc=False, _internal_bufsize=1,
                               _pass_fds={header_write_fd, sortopt_read_fd})

            # It would be nice to monitor the sort command here.  Unfortunately, there
            # is a race condition that makes this difficult.  We could loop until the
            # sort command is created, then monitor it.

        else:
            # Feed the named file into the data processing pipeline,
            input_suffix: str = input_path.suffix.lower()
            if input_suffix in [".gz", ".z"]:
                if verbose:
                    print("gunzip input file: %s" % repr(str(input_path)), file=error_file, flush=True)
                sh_gzip = sh.Command(gzip_command)
                cat_proc = sh_gzip(input_path, "-dc",
                                   _in=sys.stdin, _piped=True, _err=sys.stderr,
                                   _bg=True, _bg_exc=False, _internal_bufsize=1,
                                   _done=cat_done)

                if verbose:
                    print("full command: %s -dc %s | %s" % (gzip_command, repr(str(input_path)), cmd), file=error_file, flush=True)

            elif input_suffix in [".bz2", ".bz"]:
                if verbose:
                    print("bunzip2 input file: %s" % repr(str(input_path)), file=error_file, flush=True)
                sh_bzip2 = sh.Command(bzip2_command)
                cat_proc = sh_bzip2(input_path, "-dc",
                                    _in=sys.stdin, _piped=True, _err=sys.stderr,
                                    _bg=True, _bg_exc=False, _internal_bufsize=1,
                                    _done=cat_done)

                if verbose:
                    print("full command: %s -dc %s | %s" % (bzip2_command, repr(str(input_path)), cmd), file=error_file, flush=True)

            elif input_suffix in [".xz", ".lzma"]:
                if verbose:
                    print("unxz input file: %s" % repr(str(input_path)), file=error_file, flush=True)
                sh_xz = sh.Command(xz_command)
                cat_proc = sh_xz(input_path, "-dc",
                                 _in=sys.stdin, _piped=True, _err=sys.stderr,
                                 _bg=True, _bg_exc=False, _internal_bufsize=1,
                                 _done=cat_done)
                if verbose:
                    print("full command: %s -dc %s | %s" % (xz_command, repr(str(input_path)), cmd), file=error_file, flush=True)

            else:
                if verbose:
                    print("input file: %s" % repr(str(input_path)), file=error_file, flush=True)
                cat_proc = sh.cat(input_path, _in=sys.stdin, _piped=True, _err=sys.stderr,
                                  _bg=True, _bg_exc=False, _internal_bufsize=1,
                                  _done=cat_done)
                if verbose:
                    print("full command: cat %s | %s" % (repr(str(input_path)), cmd), file=error_file, flush=True)


            # If enabled, monitor the progress of reading the input file.
            # Since we do not have access to the pid of the sort command,
            # we cannot monitor the progress of the merge phases.
            if verbose:
                print("Monitoring the cat command (pid=%d)." % cat_proc.pid, file=error_file, flush=True)
            progress_startup(pid=cat_proc.pid)

            # Sh version 1.13 or greater is required for _pass_fds.
            sh_bash = sh.Command(bash_command)
            cmd_proc = sh_bash(cat_proc, "-c", cmd, _out=sys.stdout, _err=sys.stderr,
                               _bg=True, _bg_exc=False, _internal_bufsize=1,
                               _pass_fds={header_write_fd, sortopt_read_fd})
            # Since we do not have access to the pid of the sort command,
            # we cannot monitor the progress of the merge phases.

        if verbose:
            print("Running the sort script (pid=%d)." % cmd_proc.pid, file=error_file, flush=True)

        if verbose:
            print("Reading the KGTK input file header line with KgtkReader", file=error_file, flush=True)
        kr: KgtkReader = KgtkReader.open(Path("<%d" % header_read_fd),
                                         options=reader_options,
                                         value_options = value_options,
                                         error_file=error_file,
                                         verbose=verbose,
                                         very_verbose=very_verbose,
                                         )
        if verbose:
            print("KGTK header: %s" % " ".join(kr.column_names), file=error_file, flush=True)

        sort_options: str = ""
        if reverse_sort:
            sort_options += " --reverse"
        if numeric_sort:
            sort_options += " --numeric"

        if extra is not None and len(extra) > 0:
            sort_options += " " + extra

        # We will consume entries in reverse_columns and numeric_columns,
        # then complain if any are left over.
        if reverse_columns is not None:
            reverse_columns = reverse_columns[:] # Protect against modifying a shared list.
        if numeric_columns is not None:
            numeric_columns = numeric_columns[:] # Protect against modifying a shared list.

        column_name: str
        sort_idx: int
        if columns is not None and len(columns) > 0:
            # Process the list of column names, including splitting
            # comma-separated lists of column names.
            for column_name in columns:
                column_name_2: str
                for column_name_2 in column_name.split(","):
                    column_name_2 = column_name_2.strip()
                    if len(column_name_2) == 0:
                        continue
                    if column_name_2.isdigit():
                        sort_idx = int(column_name_2)
                        if sort_idx > len(kr.column_names):
                            kr.close()
                            cleanup()
                            raise KGTKException("Invalid column number %d (max %d)." % (sort_idx, len(kr.column_names)))
                    else:
                        if column_name_2 not in kr.column_names:
                            kr.close()
                            cleanup()
                            raise KGTKException("Unknown column_name %s" % repr(column_name_2))
                        sort_idx = kr.column_name_map[column_name_2] + 1
                    sort_options += " -k %d,%d" % (sort_idx, sort_idx)
                    if reverse_columns is not None and column_name_2 in reverse_columns:
                        sort_options += "r"
                        reverse_columns.remove(column_name_2)
                    if numeric_columns is not None and column_name_2 in numeric_columns:
                        sort_options += "n"
                        numeric_columns.remove(column_name_2)
        else:
            # TODO: support the case where the column name in reverse_columns
            # or numeric_columns is an alias of the name used in the file header.
            if kr.is_node_file:
                sort_idx = kr.id_column_idx + 1
                sort_options += " -k %d,%d" % (sort_idx, sort_idx)
                column_name = kr.column_names[kr.id_column_idx]
                if reverse_columns is not None and column_name in reverse_columns:
                    sort_options += "r"
                    reverse_columns.remove(column_name)
                if numeric_columns is not None and column_name in numeric_columns:
                    sort_options += "n"
                    numeric_columns.remove(column_name)

            elif kr.is_edge_file:
                if kr.id_column_idx >= 0:
                    sort_idx = kr.id_column_idx + 1
                    sort_options += " -k %d,%d" % (sort_idx, sort_idx)
                    column_name = kr.column_names[kr.id_column_idx]
                    if reverse_columns is not None and column_name in reverse_columns:
                        sort_options += "r"
                        reverse_columns.remove(column_name)
                    if numeric_columns is not None and column_name in numeric_columns:
                        sort_options += "n"
                        numeric_columns.remove(column_name)

                sort_idx = kr.node1_column_idx + 1
                sort_options += " -k %d,%d" % (sort_idx, sort_idx)
                column_name = kr.column_names[kr.node1_column_idx]
                if reverse_columns is not None and column_name in reverse_columns:
                    sort_options += "r"
                    reverse_columns.remove(column_name)
                if numeric_columns is not None and column_name in numeric_columns:
                    sort_options += "n"
                    numeric_columns.remove(column_name)

                sort_idx = kr.label_column_idx + 1
                sort_options += " -k %d,%d" % (sort_idx, sort_idx)
                column_name = kr.column_names[kr.label_column_idx]
                if reverse_columns is not None and column_name in reverse_columns:
                    sort_options += "r"
                    reverse_columns.remove(column_name)
                if numeric_columns is not None and column_name in numeric_columns:
                    sort_options += "n"
                    numeric_columns.remove(column_name)

                sort_idx = kr.node2_column_idx + 1
                sort_options += " -k %d,%d" % (sort_idx, sort_idx)
                column_name = kr.column_names[kr.node2_column_idx]
                if reverse_columns is not None and column_name in reverse_columns:
                    sort_options += "r"
                    reverse_columns.remove(column_name)
                if numeric_columns is not None and column_name in numeric_columns:
                    numeric_columns.remove(column_name)
                    sort_options += "n"

            else:
                cleanup()
                raise KGTKException("Unknown KGTK file mode, please specify the sorting columns.")

        # Check for unconsumed entries in reverse_columns and numeric_columns:
        if reverse_columns is not None and len(reverse_columns) > 0:
            raise KGTKException("Unknown reverse column(s) %s" % " ".join([repr(column_name) for column_name in reverse_columns]))
        if numeric_columns is not None and len(numeric_columns) > 0:
            raise KGTKException("Unknown numeric column(s) %s" % " ".join([repr(column_name) for column_name in numeric_columns]))

        if verbose:
            print("sort options: %s" % sort_options, file=error_file, flush=True)

        kr.close() # We are done with the KgtkReader now.

        # Send the sort options back to the data processing pipeline.
        with open(sortopt_write_fd, "w") as options_file:
            options_file.write(sort_options + "\n")

        if verbose:
            print("\nWaiting for the sort command to complete.\n", file=error_file, flush=True)
        cmd_proc.wait()

        if verbose:
            print("Cleanup.", file=error_file, flush=True)
        cleanup()

        return 0

    except Exception as e:
        # import traceback
        # traceback.print_tb(sys.exc_info()[2], 10)
        raise KGTKException('INTERNAL ERROR: ' + type(e).__module__ + '.' + str(e) + '\n')




"""
This module updates files with a version string.

 * The version string contains a timestamp and a digest.
   * The digest identifies the content of the file.
     * Calcuated using blake2b
   * The timestamp indicates when a change was last detected.
 * The digest is calculated on the input file after removing
the version string from the input file.
   * If a version string is not detected in the input file,
     no action is taken on that file.
 * If the computed digest of the input file is different from
   the recorded digest:
   * The input file is updated with a new version string.
   * The input file is rewritten its filesystem.
"""

from argparse import ArgumentParser
import attr
import base64
import datetime as dt
import hashlib
from pathlib import Path
import re
import sys
import typing

from kgtk.utils.argparsehelpers import optional_bool

@attr.s(slots=True, frozen=True)
class UpdateVersion():
    UPDATE_VERSION: str = "2020-07-22T22:10:11.047901+00:00#uwVDNJhT34iJFhsPf5vafIhNdaD3xcZ0RxgPN9CPRmYY7cfpS5guHbAY7hApoK1aue3W+cz9ftP8kXJIK+budQ=="
    version_pattern: str = attr.ib(validator=attr.validators.instance_of(str), default=r'^\s*UPDATE_VERSION\s*:\s*str\s*=\s*"(?P<version>.*)"$')
    blake_bloke: str = attr.ib(validator=attr.validators.instance_of(str), default="UpdateVersionHsh")

    allow_updates: bool = attr.ib(validator=attr.validators.instance_of(bool), default=True)
    show_changes: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    # Feedback and error output:
    error_file: typing.TextIO = attr.ib(default=sys.stderr)
    verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    def process_file(self, filepath: Path)->bool:
        hasher = hashlib.blake2b(person=self.blake_bloke.encode())
        version_pattern_re = re.compile(self.version_pattern.encode())

        current_version: typing.Optional[bytes] = None

        lines: typing.List[bytes] = []
        line: bytes
            
        if self.verbose:
            print("Checking the version of %s" % str(filepath), file=self.error_file, flush=True)
        with open(filepath, "rb") as ifile:
            for line in ifile:
                lines.append(line)
                matches = version_pattern_re.match(line)
                if matches:
                    # Retain the last match if multiple hits.
                    current_version = matches.group("version")
                    line = re.sub(rb'".*"', b'""', line)
                hasher.update(line)

        if current_version is None:
            if self.verbose:
                print("No version mark found in %s" % str(filepath))
            return False
        current_timestamp: bytes = b''
        current_digest: bytes = b''
        if len(current_version) > 0:
            current_timestamp, current_digest = current_version.split(b'#', 1)

        new_digest: bytes = base64.b64encode(hasher.digest())
        if new_digest == current_digest:
            if self.verbose:
                print("%s has not changed since %s" % (str(filepath), str(current_timestamp, "utf-8")), file=self.error_file, flush=True)
            return False

        if not self.allow_updates:
            if self.verbose:
                print("%s has changed since %s, but updates are disabled." % (str(filepath), str(current_timestamp, "utf-8")), file=self.error_file, flush=True)
            return True

        new_timestamp: bytes = dt.datetime.now(dt.timezone.utc).isoformat().encode()
        new_version: bytes = new_timestamp + b'#' + new_digest
        
        if self.verbose or self.show_changes:
            print("Updating the version mark in %s" % str(filepath), file=self.error_file, flush=True)
        with open(filepath, "wb") as ofile:
            for line in lines:
                matches = version_pattern_re.match(line)
                if matches:
                    line = re.sub(rb'".*"', b'"' + new_version + b'"', line)
                ofile.write(line)

        return True

    def process_files(self, filepaths: typing.List[Path])->int:
        """
        Returns the count of files that changed.
        """
        changes: int = 0
        filepath: Path
        for filepath in filepaths:
            if self.process_file(filepath):
                changes += 1


        if self.verbose:
            print("%d files updated" % changes, file=self.error_file, flush=True)

        return changes

def main():
    """
    Update version marks.
    """
    parser = ArgumentParser()
    parser.add_argument(dest="filepaths", help="The file(s) to process", type=Path, nargs="+")

    parser.add_argument(       "--allow-updates", dest="allow_updates", type=optional_bool, nargs='?', const=True, default=True,
                               help="Allow updates. (default=%(default)s).")
    
    parser.add_argument(       "--show-changes", dest="show_changes", type=optional_bool, nargs='?', const=True, default=True,
                               help="Write a line for each changed file. (default=%(default)s).")
    
    parser.add_argument(       "--show-version", dest="show_version", type=optional_bool, nargs='?', const=True, default=False,
                               help="Print the version of this program. (default=%(default)s).")

    parser.add_argument("-v", "--verbose", dest="verbose", type=optional_bool, nargs='?', const=True, default=False,
                        help="Print additional progress messages (default=%(default)s).")
    
    args: Namespace = parser.parse_args()

    if args.show_version:
        print("Version: %s" % UpdateVersion.UPDATE_VERSION)

    updater: UpdateVersion = UpdateVersion(allow_updates=args.allow_updates,
                                           show_changes=args.show_changes,
                                           verbose=args.verbose)
    updater.process_files(args.filepaths)
    
if __name__ == "__main__":
    main()

